#  [ANDES] (C)2015-2022 Hantao Cui
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#

"""
Module for ANDES models.
"""

import logging
import pprint
from collections import OrderedDict
from textwrap import wrap
from typing import Callable, Iterable, Union

import numpy as np
import scipy as sp
from andes.core.block import Block
from andes.core.common import Config, JacTriplet, ModelFlags
from andes.core.discrete import Discrete
from andes.core.documenter import Documenter
from andes.core.model.modelcache import ModelCache
from andes.core.model.modelcall import ModelCall
from andes.core.param import ExtParam
from andes.core.service import (ApplyFunc, BackRef, BaseService, ConstService,
                                DeviceFinder, ExtService, FlagValue,
                                InitChecker, NumReduce, NumRepeat, NumSelect,
                                ParamCalc, PostInitService, RandomService,
                                Replace, SubsService, SwBlock, VarService)
from andes.core.symprocessor import SymProcessor
from andes.core.var import Algeb, BaseVar, ExtAlgeb, ExtState, State
from andes.shared import jac_full_names, numba
from andes.utils.func import list_flatten
from andes.utils.tab import Tab

logger = logging.getLogger(__name__)


class Model:
    r"""
    Base class for power system DAE models.


    After subclassing `ModelData`, subclass `Model`` to complete a DAE model.
    Subclasses of `Model` define DAE variables, services, and other types of parameters,
    in the constructor ``__init__``.

    Attributes
    ----------
    num_params : OrderedDict
        {name: instance} of numerical parameters, including internal
        and external ones

    Examples
    --------
    Take the static PQ as an example, the subclass of `Model`, `PQ`, should look like ::

        class PQ(PQData, Model):
            def __init__(self, system, config):
                PQData.__init__(self)
                Model.__init__(self, system, config)

    Since `PQ` is calling the base class constructors, it is meant to be the final class and not further
    derived.
    It inherits from `PQData` and `Model` and must call constructors in the order of `PQData` and `Model`.
    If the derived class of `Model` needs to be further derived, it should only derive from `Model`
    and use a name ending with `Base`. See :py:class:`andes.models.synchronous.GENBASE`.

    Next, in `PQ.__init__`, set proper flags to indicate the routines in which the model will be used ::

        self.flags.update({'pflow': True})

    Currently, flags `pflow` and `tds` are supported. Both are `False` by default, meaning the model is
    neither used in power flow nor time-domain simulation. **A very common pitfall is forgetting to set the flag**.

    Next, the group name can be provided. A group is a collection of models with common parameters and variables.
    Devices' idx of all models in the same group must be unique. To provide a group name, use ::

        self.group = 'StaticLoad'

    The group name must be an existing class name in :py:mod:`andes.models.group`.
    The model will be added to the specified group and subject to the variable and parameter policy of the
    group.
    If not provided with a group class name, the model will be placed in the `Undefined` group.

    Next, additional configuration flags can be added.
    Configuration flags for models are load-time variables, specifying the behavior of a model.
    They can be exported to an `andes.rc` file and automatically loaded when creating the `System`.
    Configuration flags can be used in equation strings, as long as they are numerical values.
    To add config flags, use ::

        self.config.add(OrderedDict((('pq2z', 1), )))

    It is recommended to use `OrderedDict` instead of `dict`, although the syntax is verbose.
    Note that booleans should be provided as integers (1 or 0), since `True` or `False` is interpreted as
    a string when loaded from the `rc` file and will cause an error.

    Next, it's time for variables and equations! The `PQ` class does not have internal variables itself.
    It uses its `bus` parameter to fetch the corresponding `a` and `v` variables of buses.
    Equation wise, it imposes an active power and a reactive power load equation.

    To define external variables from `Bus`, use ::

            self.a = ExtAlgeb(model='Bus', src='a',
                              indexer=self.bus, tex_name=r'\theta')
            self.v = ExtAlgeb(model='Bus', src='v',
                              indexer=self.bus, tex_name=r'V')

    Refer to the subsection Variables for more details.

    The simplest `PQ` model will impose constant P and Q, coded as ::

            self.a.e_str = "u * p"
            self.v.e_str = "u * q"

    where the `e_str` attribute is the equation string attribute. `u` is the connectivity status.
    Any parameter, config, service or variable can be used in equation strings.

    Three additional scalars can be used in equations:
    - ``dae_t`` for the current simulation time (can be used if the model has flag `tds`).
    - ``sys_f`` for system frequency (from ``system.config.freq``).
    - ``sys_mva`` for system base mva (from ``system.config.mva``).

    The above example is overly simplified. Our `PQ` model wants a feature to switch itself to
    a constant impedance if the voltage is out of the range `(vmin, vmax)`.
    To implement this, we need to introduce a discrete component called `Limiter`, which yields three arrays
    of binary flags, `zi`, `zl`, and `zu` indicating in-range, below lower-limit, and above upper-limit,
    respectively.

    First, create an attribute `vcmp` as a `Limiter` instance ::

            self.vcmp = Limiter(u=self.v, lower=self.vmin, upper=self.vmax,
                                 enable=self.config.pq2z)

    where `self.config.pq2z` is a flag to turn this feature on or off.
    After this line, we can use `vcmp_zi`, `vcmp_zl`, and `vcmp_zu` in other equation strings. ::

            self.a.e_str = "u * (p0 * vcmp_zi + " \
                           "p0 * vcmp_zl * (v ** 2 / vmin ** 2) + " \
                           "p0 * vcmp_zu * (v ** 2 / vmax ** 2))"

            self.v.e_str = "u * (q0 * vcmp_zi + " \
                           "q0 * vcmp_zl * (v ** 2 / vmin ** 2) + "\
                           "q0 * vcmp_zu * (v ** 2 / vmax ** 2))"

    Note that `PQ.a.e_str` can use the three variables from `vcmp` even before defining `PQ.vcmp`, as long as
    `PQ.vcmp` is defined, because `vcmp_zi` is just a string literal in `e_str`.

    The two equations above implement a piece-wise power injection equation. It selects the original power demand
    if within range, and uses the calculated power when out of range.

    Finally, to let ANDES pick up the model, the model name needs to be added to `models/__init__.py`.
    Follow the examples in the `OrderedDict`, where the key is the file name, and the value is the class name.

    Notes
    -----
    To modify parameters or services use ``set()``, which writes directly to the given attribute,
    or ``alter()``, which converts parameters to system base like that for input data.

    """

    def __init__(self, system=None, config=None):
        self.system = system

        # duplicate attributes from ModelData. Keep for now.
        self.n = 0
        self.group = 'Undefined'

        # params and vars that exist in the group but not in this model
        # normally empty but can be used in special cases to bypass
        # shared param/var checking
        self.group_param_exception = list()
        self.group_var_exception = list()

        if not hasattr(self, 'num_params'):
            self.num_params = OrderedDict()
        if not hasattr(self, 'cache'):
            self.cache = ModelCache()

        # variables
        self.states = OrderedDict()  # internal states
        self.states_ext = OrderedDict()  # external states
        self.algebs = OrderedDict()  # internal algebraic variables
        self.algebs_ext = OrderedDict()  # external algebraic vars
        self.vars_decl_order = OrderedDict()  # variable in the order of declaration

        self.params_ext = OrderedDict()  # external parameters

        self.discrete = OrderedDict()  # discrete comp.
        self.blocks = OrderedDict()  # blocks

        self.services = OrderedDict()  # service/temporary variables
        self.services_var = OrderedDict()  # variable services updated each step/iter
        self.services_var_seq = OrderedDict()
        self.services_var_nonseq = OrderedDict()
        self.services_post = OrderedDict()  # post-initialization storage services
        self.services_subs = OrderedDict()  # to-be-substituted services
        self.services_icheck = OrderedDict()  # post-initialization check services
        self.services_ref = OrderedDict()  # BackRef
        self.services_fnd = OrderedDict()  # services to find/add devices
        self.services_ext = OrderedDict()  # external services (to be retrieved)
        self.services_ops = OrderedDict()  # operational services (for special usages)

        self.tex_names = OrderedDict((('dae_t', 't_{dae}'),
                                      ('sys_f', 'f_{sys}'),
                                      ('sys_mva', 'S_{b,sys}'),
                                      ))

        # Model behavior flags
        self.flags = ModelFlags()

        # `in_use` is used by models with `BackRef` when not reference
        self.in_use = True  # True if this model is in use, False removes this model from all calls

        self.config = Config(name=self.class_name)  # `config` that can be exported
        if config is not None:
            self.config.load(config)

        # basic configs
        self.config.add(OrderedDict((('allow_adjust', 1),
                                    ('adjust_lower', 0),
                                    ('adjust_upper', 1),
                                     )))

        self.config.add_extra("_help",
                              allow_adjust='allow adjusting upper or lower limits',
                              adjust_lower='adjust lower limit',
                              adjust_upper='adjust upper limit',
                              )

        self.config.add_extra("_alt",
                              allow_adjust=(0, 1),
                              adjust_lower=(0, 1),
                              adjust_upper=(0, 1),
                              )

        self.calls = ModelCall()  # callback and LaTeX string storage
        self.triplets = JacTriplet()  # Jacobian triplet storage
        self.syms = SymProcessor(self)  # symbolic processor instance
        self.docum = Documenter(self)

        # cached class attributes
        self.cache.add_callback('all_vars', self._all_vars)
        self.cache.add_callback('iter_vars', self._iter_vars)
        self.cache.add_callback('input_vars', self._input_vars)
        self.cache.add_callback('output_vars', self._output_vars)

        self.cache.add_callback('all_vars_names', self._all_vars_names)
        self.cache.add_callback('all_params', self._all_params)
        self.cache.add_callback('all_params_names', self._all_params_names)
        self.cache.add_callback('algebs_and_ext', self._algebs_and_ext)
        self.cache.add_callback('states_and_ext', self._states_and_ext)
        self.cache.add_callback('services_and_ext', self._services_and_ext)
        self.cache.add_callback('vars_ext', self._vars_ext)
        self.cache.add_callback('vars_int', self._vars_int)
        self.cache.add_callback('v_getters', self._v_getters)
        self.cache.add_callback('v_adders', self._v_adders)
        self.cache.add_callback('v_setters', self._v_setters)
        self.cache.add_callback('e_adders', self._e_adders)
        self.cache.add_callback('e_setters', self._e_setters)

        self._input = OrderedDict()  # cached dictionary of inputs
        self._input_z = OrderedDict()  # discrete flags, storage only.
        self._rhs_f = OrderedDict()  # RHS of external f
        self._rhs_g = OrderedDict()  # RHS of external g

        self.f_args = []
        self.g_args = []  # argument value lists
        self.j_args = dict()
        self.s_args = OrderedDict()
        self.ia_args = OrderedDict()
        self.ii_args = OrderedDict()
        self.ij_args = OrderedDict()

        self.coeffs = dict()  # pu conversion coefficient storage
        self.bases = dict()   # base storage, such as Vn, Vb, Zn, Zb
        self.debug_equations = list()  # variable names for debugging corresponding equation

    def _register_attribute(self, key, value):
        """
        Register a pair of attributes to the model instance.

        Called within ``__setattr__``, this is where the magic happens.
        Subclass attributes are automatically registered based on the variable type.
        Block attributes will be exported and registered recursively.
        """
        if isinstance(value, Algeb):
            self.algebs[key] = value
        elif isinstance(value, ExtAlgeb):
            self.algebs_ext[key] = value
        elif isinstance(value, State):
            self.states[key] = value
        elif isinstance(value, ExtState):
            self.states_ext[key] = value
        elif isinstance(value, ExtParam):
            self.params_ext[key] = value
        elif isinstance(value, Discrete):
            self.discrete[key] = value
        elif isinstance(value, ConstService):  # services with only `v_str`
            self.services[key] = value
            # store VarService in an additional dict
            if isinstance(value, VarService):
                self.services_var[key] = value
                if value.sequential:
                    self.services_var_seq[key] = value
                else:
                    self.services_var_nonseq[key] = value
            elif isinstance(value, PostInitService):
                self.services_post[key] = value
        elif isinstance(value, SubsService):
            self.services_subs[key] = value
        elif isinstance(value, DeviceFinder):
            self.services_fnd[key] = value
        elif isinstance(value, BackRef):
            self.services_ref[key] = value
        elif isinstance(value, ExtService):
            self.services_ext[key] = value
        elif isinstance(value, (NumRepeat, NumReduce, NumSelect,
                                FlagValue, RandomService,
                                SwBlock,
                                ParamCalc, Replace, ApplyFunc)):
            self.services_ops[key] = value
        elif isinstance(value, InitChecker):
            self.services_icheck[key] = value
        elif isinstance(value, Block):
            self.blocks[key] = value
            # pull in sub-variables from control blocks
            if value.namespace == 'local':
                prepend = value.name + '_'
                tex_append = value.tex_name
            else:
                prepend = ''
                tex_append = ''

            for var_name, var_instance in value.export().items():
                var_instance.name = f'{prepend}{var_name}'
                var_instance.tex_name = f'{var_instance.tex_name}_{{{tex_append}}}'
                self.__setattr__(var_instance.name, var_instance)

    def _check_attribute(self, key, value):
        """
        Check the attribute pair for valid names while instantiating the class.

        This function assigns `owner` to the model itself, assigns the name and tex_name.
        """
        if isinstance(value, (BaseVar, BaseService, Discrete, Block)):
            if not value.owner:
                value.owner = self
            if not value.name:
                value.name = key
            if not value.tex_name:
                value.tex_name = key
            if key in self.__dict__:
                logger.warning(f"{self.class_name}: redefinition of member <{key}>. Likely a modeling error.")

    def __setattr__(self, key, value):
        """
        Overload the setattr function to register attributes.

        Parameters
        ----------
        key : str
            name of the attribute
        value : [BaseVar, BaseService, Discrete, Block]
            value of the attribute
        """

        self._check_attribute(key, value)

        # store the variable declaration order
        if isinstance(value, BaseVar):
            value.id = len(self._all_vars())  # NOT in use yet
            self.vars_decl_order[key] = value

        self._register_attribute(key, value)

        super(Model, self).__setattr__(key, value)

    def idx2uid(self, idx):
        """
        Convert idx to the 0-indexed unique index.

        Parameters
        ----------
        idx : array-like, numbers, or str
            idx of devices

        Returns
        -------
        list
            A list containing the unique indices of the devices
        """
        if idx is None:
            logger.debug("idx2uid returned None for idx None")
            return None
        if isinstance(idx, (float, int, str, np.integer, np.floating)):
            return self._one_idx2uid(idx)
        elif isinstance(idx, Iterable):
            if len(idx) > 0 and isinstance(idx[0], (list, np.ndarray)):
                idx = list_flatten(idx)
            return [self._one_idx2uid(i) if i is not None else None
                    for i in idx]
        else:
            raise NotImplementedError(f'Unknown idx type {type(idx)}')

    def _one_idx2uid(self, idx):
        """
        Helper function for checking if an idx exists and
        converting it to uid.
        """

        if idx not in self.uid:
            raise KeyError("<%s>: device not exist with idx=%s." %
                           (self.class_name, idx))

        return self.uid[idx]

    def set_backref(self, name, from_idx, to_idx):
        """
        Helper function for setting idx-es to ``BackRef``.
        """

        if name not in self.services_ref:
            return

        uid = self.idx2uid(to_idx)
        self.services_ref[name].v[uid].append(from_idx)

    def get(self, src: str, idx, attr: str = 'v', allow_none=False, default=0.0):
        """
        Get the value of an attribute of a model property.

        The return value is ``self.<src>.<attr>[idx]``

        Parameters
        ----------
        src : str
            Name of the model property
        idx : str, int, float, array-like
            Indices of the devices
        attr : str, optional, default='v'
            The attribute of the property to get.
            ``v`` for values, ``a`` for address, and ``e`` for equation value.
        allow_none : bool
            True to allow None values in the indexer
        default : float
            If `allow_none` is true, the default value to use for None indexer.

        Returns
        -------
        array-like
            ``self.<src>.<attr>[idx]``

        """
        uid = self.idx2uid(idx)
        if isinstance(self.__dict__[src].__dict__[attr], list):
            if isinstance(uid, Iterable):
                if not allow_none and (uid is None or None in uid):
                    raise KeyError('None not allowed in uid/idx. Enable through '
                                   '`allow_none` and provide a `default` if needed.')
                return [self.__dict__[src].__dict__[attr][i] if i is not None else default
                        for i in uid]

        return self.__dict__[src].__dict__[attr][uid]

    def set(self, src, idx, attr, value):
        """
        Set the value of an attribute of a model property.

        Performs ``self.<src>.<attr>[idx] = value``.

        Parameters
        ----------
        src : str
            Name of the model property
        idx : str, int, float, array-like
            Indices of the devices
        attr : str, optional, default='v'
            The internal attribute of the property to get.
            ``v`` for values, ``a`` for address, and ``e`` for equation value.
        value : array-like
            New values to be set

        Returns
        -------
        bool
            True when successful.
        """
        uid = self.idx2uid(idx)
        self.__dict__[src].__dict__[attr][uid] = value
        return True

    def alter(self, src, idx, value):
        """
        Alter input parameter or service values.

        If operates on a parameter, the input should be in the same
        base as that in the input file. This function will convert
        the new value to system-base per unit.

        Parameters
        ----------
        src : str
            The parameter name to alter
        idx : str, float, int
            The device to alter
        value : float
            The desired value
        """
        instance = self.__dict__[src]

        if hasattr(instance, 'vin') and (instance.vin is not None):
            self.set(src, idx, 'vin', value)
            instance.v[:] = instance.vin * instance.pu_coeff
        else:
            self.set(src, idx, 'v', value)

    def get_inputs(self, refresh=False):
        """
        Get an OrderedDict of the inputs to the numerical function calls.

        Parameters
        ----------
        refresh : bool
            Refresh the values in the dictionary.
            This is only used when the memory addresses of arrays change.
            After initialization, all array assignments are in place.
            To avoid overhead, refresh should not be used after initialization.

        Returns
        -------
        OrderedDict
            The input name and value array pairs in an OrderedDict

        Notes
        -----
        `dae.t` is now a numpy.ndarray which has stable memory.
        There is no need to refresh `dat_t` in this version.

        """
        if len(self._input) == 0 or refresh:
            self.refresh_inputs()
            self.refresh_inputs_arg()

        return self._input

    def refresh_inputs(self):
        """
        This is the helper function to refresh inputs.

        The functions collects object references into ``OrderedDict``
        `self._input` and `self._input_z`.

        Returns
        -------
        None

        """
        # The order of inputs: `all_params` and then `all_vars`, finally `config`
        # the below sequence should correspond to `self.cache.all_params_names`
        for instance in self.num_params.values():
            self._input[instance.name] = instance.v

        for instance in self.services.values():
            self._input[instance.name] = instance.v

        for instance in self.services_ext.values():
            self._input[instance.name] = instance.v

        for instance in self.services_ops.values():
            self._input[instance.name] = instance.v

        # discrete flags
        for instance in self.discrete.values():
            for name, val in zip(instance.get_names(), instance.get_values()):
                self._input[name] = val
                self._input_z[name] = val

        # append all variable values
        for instance in self.cache.all_vars.values():
            self._input[instance.name] = instance.v

        # append config variables as arrays
        for key, val in self.config.as_dict(refresh=True).items():
            self._input[key] = np.array(val)

        # zeros and ones
        self._input['__zeros'] = np.zeros(self.n)
        self._input['__ones'] = np.ones(self.n)
        self._input['__falses'] = np.full(self.n, False)
        self._input['__trues'] = np.full(self.n, True)

        # --- below are numpy scalars ---
        # update`dae_t` and `sys_f`, and `sys_mva`
        self._input['sys_f'] = np.array(self.system.config.freq, dtype=float)
        self._input['sys_mva'] = np.array(self.system.config.mva, dtype=float)
        self._input['dae_t'] = self.system.dae.t

    def mock_refresh_inputs(self):
        """
        Use mock data to fill the inputs.

        This function is used to generate input data of the desired type
        to trigget JIT compilation.
        """

        self.get_inputs()
        mock_arr = np.array([1.])

        for key in self._input.keys():
            try:
                key_ndim = self._input[key].ndim
            except AttributeError:
                logger.error(key)
                logger.error(self.class_name)

            key_type = self._input[key].dtype

            if key_ndim == 0:
                self._input[key] = mock_arr.reshape(()).astype(key_type)
            elif key_ndim == 1:
                self._input[key] = mock_arr.astype(key_type)

            else:
                raise NotImplementedError("Unkonwn input data dimension %s" % key_ndim)

    def refresh_inputs_arg(self):
        """
        Refresh inputs for each function with individual argument list.
        """
        self.f_args = list()
        self.g_args = list()
        self.j_args = dict()
        self.s_args = dict()
        self.ii_args = dict()
        self.ia_args = dict()
        self.ij_args = dict()

        self.f_args = [self._input[arg] for arg in self.calls.f_args]
        self.g_args = [self._input[arg] for arg in self.calls.g_args]
        self.sns_args = [self._input[arg] for arg in self.calls.sns_args]

        # each value below is a dict
        mapping = {
            'j_args': self.j_args,
            's_args': self.s_args,
            'ia_args': self.ia_args,
            'ii_args': self.ii_args,
            'ij_args': self.ij_args,
        }

        for key, val in mapping.items():
            source = self.calls.__dict__[key]
            for name in source:
                val[name] = [self._input[arg] for arg in source[name]]

    def l_update_var(self, dae_t, *args, niter=None, err=None, **kwargs):
        """
        Call the ``check_var`` method of discrete components to update the internal status flags.

        The function is variable-dependent and should be called before updating equations.

        Returns
        -------
        None
        """
        for instance in self.discrete.values():
            if instance.has_check_var:
                instance.check_var(dae_t=dae_t, niter=niter, err=err)

    def l_check_eq(self, init=False, **kwargs):
        """
        Call the ``check_eq`` method of discrete components to update equation-dependent flags.

        This function should be called after equation updates.
        AntiWindup limiters use it to append pegged states to the ``x_set`` list.

        Returns
        -------
        None
        """
        if init:
            for instance in self.discrete.values():
                if instance.has_check_eq:
                    instance.check_eq(allow_adjust=self.config.allow_adjust,
                                      adjust_lower=self.config.adjust_lower,
                                      adjust_upper=self.config.adjust_upper,
                                      )
        else:
            for instance in self.discrete.values():
                if instance.has_check_eq:
                    instance.check_eq()

    def s_update(self):
        """
        Update service equation values.

        This function is only evaluated at initialization. Service values are
        updated sequentially. The ``v`` attribute of services will be assigned
        at a new memory address.
        """
        for name, instance in self.services.items():
            if name in self.calls.s:
                func = self.calls.s[name]
                if callable(func):
                    self.get_inputs(refresh=True)
                    # NOTE:
                    # Use new assignment due to possible size change.
                    # Always make a copy and make the RHS a 1-d array
                    instance.v = np.ravel(np.array(func(*self.s_args[name]), dtype=instance.vtype))
                else:
                    instance.v = np.ravel(np.array(func, dtype=instance.vtype))

                # convert to an array if the return of lambda function is a scalar
                if isinstance(instance.v, (int, float)):
                    instance.v = np.ones(self.n, dtype=instance.vtype) * instance.v
                elif isinstance(instance.v, np.ndarray) and len(instance.v) == 1:
                    instance.v = np.ones(self.n, dtype=instance.vtype) * instance.v

            # --- Very Important ---
            # the numerical call of a `ConstService` should only depend on previously
            #   evaluated variables.
            func = instance.v_numeric
            if func is not None and callable(func):
                kwargs = self.get_inputs(refresh=True)
                instance.v = func(**kwargs).copy().astype(instance.vtype)  # performs type conv.

        if self.flags.s_num is True:
            kwargs = self.get_inputs(refresh=True)
            self.s_numeric(**kwargs)

        # Block-level `s_numeric` not supported.
        self.get_inputs(refresh=True)

    def s_update_var(self):
        """
        Update values of :py:class:`andes.core.service.VarService`.
        """

        if len(self.services_var):
            kwargs = self.get_inputs()
            # evaluate `v_str` functions for sequential VarService
            for name, instance in self.services_var_seq.items():
                if instance.v_str is not None:
                    func = self.calls.s[name]
                    if callable(func):
                        instance.v[:] = func(*self.s_args[name])

            # apply non-sequential services from ``v_str``:w
            if callable(self.calls.sns):
                ret = self.calls.sns(*self.sns_args)
                for idx, instance in enumerate(self.services_var_nonseq.values()):
                    instance.v[:] = ret[idx]

            # Apply individual `v_numeric`
            for instance in self.services_var.values():
                if instance.v_numeric is None:
                    continue
                if callable(instance.v_numeric):
                    instance.v[:] = instance.v_numeric(**kwargs)

        if self.flags.sv_num is True:
            kwargs = self.get_inputs()
            self.s_numeric_var(**kwargs)

        # Block-level `s_numeric_var` not supported.

    def s_update_post(self):
        """
        Update post-initialization services.
        """

        kwargs = self.get_inputs()
        if len(self.services_post):
            for name, instance in self.services_post.items():
                func = self.calls.s[name]
                if callable(func):
                    instance.v[:] = func(*self.s_args[name])

            for instance in self.services_post.values():
                func = instance.v_numeric
                if func is not None and callable(func):
                    instance.v[:] = func(**kwargs)

    def _init_wrap(self, x0, params):
        """
        A wrapper for converting the initialization equations into standard forms g(x) = 0, where x is an array.
        """
        vars_input = []
        for i, _ in enumerate(self.cache.iter_vars.values()):
            vars_input.append(x0[i * self.n: (i + 1) * self.n])

        ret = np.ravel(self.calls.init_std(vars_input, params))
        return ret

    def store_sparse_pattern(self):
        """
        Store rows and columns of the non-zeros in the Jacobians for building the sparsity pattern.

        This function converts the internal 0-indexed equation/variable address to the numerical addresses for
        the loaded system.

        Calling sequence:
        For each Jacobian name, `fx`, `fy`, `gx` and `gy`, store by
        a) generated constant and variable Jacobians
        c) user-provided constant and variable Jacobians,
        d) user-provided block constant and variable Jacobians

        Notes
        -----
        If `self.n == 0`, skipping this function will avoid appending empty lists/arrays and
        non-empty values, which, as a combination, is not accepted by `kvxopt.spmatrix`.
        """

        self.triplets.clear_ijv()
        if self.n == 0:  # do not check `self.in_use` here
            return

        if self.flags.address is False:
            return

        # store model-level user-defined Jacobians
        if self.flags.j_num is True:
            self.j_numeric()

        # store and merge user-defined Jacobians in blocks
        for instance in self.blocks.values():
            if instance.flags.j_num is True:
                instance.j_numeric()
                self.triplets.merge(instance.triplets)

        # for all combinations of Jacobian names (fx, fxc, gx, gxc, etc.)
        for j_name in jac_full_names:
            for idx, val in enumerate(self.calls.vjac[j_name]):

                row_name, col_name = self._jac_eq_var_name(j_name, idx)
                row_idx = self.__dict__[row_name].a
                col_idx = self.__dict__[col_name].a

                if len(row_idx) != len(col_idx):
                    logger.error(f'row {row_name}, row_idx: {row_idx}')
                    logger.error(f'col {col_name}, col_idx: {col_idx}')
                    raise ValueError(f'{self.class_name}: non-matching row_idx and col_idx')

                # Note:
                # n_elem: number of elements in the equation or variable
                # It does not necessarily equal to the number of devices of the model
                # For example, `COI.omega_sub.n` depends on the number of generators linked to the COI
                # and is likely different from `COI.n`

                n_elem = self.__dict__[row_name].n

                if j_name[-1] == 'c':
                    value = val * np.ones(n_elem)
                else:
                    value = np.zeros(n_elem)

                self.triplets.append_ijv(j_name, row_idx, col_idx, value)

    def _jac_eq_var_name(self, j_name, idx):
        """
        Get the equation and variable name for a Jacobian type based on the absolute index.
        """
        var_names_list = list(self.cache.all_vars.keys())

        eq_names = {'f': var_names_list[:len(self.cache.states_and_ext)],
                    'g': var_names_list[len(self.cache.states_and_ext):]}

        row = self.calls.ijac[j_name][idx]
        col = self.calls.jjac[j_name][idx]

        try:
            row_name = eq_names[j_name[0]][row]  # where jname[0] is the equation name (f, g, r, t)
            col_name = var_names_list[col]
        except IndexError as e:
            logger.error("Generated code outdated. Run `andes prepare -i` to re-generate.")
            raise e

        return row_name, col_name

    def get_init_order(self):
        """
        Get variable initialization order and send to `logger.info`.
        """
        out = []
        for name in self.vars_decl_order.keys():
            out.append(name)

        logger.info(f'Initialization order: \n{", ".join(out)}')

    def f_update(self):
        """
        Evaluate differential equations.

        Notes
        -----
        In-place equations: added to the corresponding DAE array.
        Non-in-place equations: in-place set to internal array to
        overwrite old values (and avoid clearing).
        """
        if callable(self.calls.f):
            f_ret = self.calls.f(*self.f_args)
            for i, var in enumerate(self.cache.states_and_ext.values()):
                if var.e_inplace:
                    var.e += f_ret[i]
                else:
                    var.e[:] = f_ret[i]

        kwargs = self.get_inputs()
        # user-defined numerical calls defined in the model
        if self.flags.f_num is True:
            self.f_numeric(**kwargs)

        # user-defined numerical calls in blocks
        for instance in self.blocks.values():
            if instance.flags.f_num is True:
                instance.f_numeric(**kwargs)

    def g_update(self):
        """
        Evaluate algebraic equations.
        """
        if callable(self.calls.g):
            g_ret = self.calls.g(*self.g_args)
            for i, var in enumerate(self.cache.algebs_and_ext.values()):
                if var.e_inplace:
                    var.e += g_ret[i]
                else:
                    var.e[:] = g_ret[i]

        kwargs = self.get_inputs()
        # numerical calls defined in the model
        if self.flags.g_num is True:
            self.g_numeric(**kwargs)

        # numerical calls in blocks
        for instance in self.blocks.values():
            if instance.flags.g_num is True:
                instance.g_numeric(**kwargs)

    def j_update(self):
        """
        Update Jacobian elements.

        Values are stored to ``Model.triplets[jname]``, where ``jname`` is a jacobian name.

        Returns
        -------
        None
        """
        for jname, jfunc in self.calls.j.items():
            ret = jfunc(*self.j_args[jname])

            for idx, fun in enumerate(self.calls.vjac[jname]):
                try:
                    self.triplets.vjac[jname][idx][:] = ret[idx]
                except (ValueError, IndexError, FloatingPointError) as e:
                    row_name, col_name = self._jac_eq_var_name(jname, idx)
                    logger.error('%s: error calculating or storing Jacobian <%s>: j_idx=%s, d%s / d%s',
                                 self.class_name, jname, idx, row_name, col_name)

                    raise e

    def get_times(self):
        """
        Get event switch_times from `TimerParam`.

        Returns
        -------
        list
            A list containing all switching times defined in TimerParams
        """
        out = []
        if self.n > 0:
            for instance in self.timer_params.values():
                out.append(instance.v)

        return out

    def switch_action(self, dae_t):
        """
        Call the switch actions.

        Parameters
        ----------
        dae_t : float
            Current simulation time

        Returns
        -------
        None

        Warnings
        --------
        Timer exported from blocks are supposed to work
        but have not been tested.
        """
        for timer in self.timer_params.values():
            if timer.callback is not None:
                timer.callback(timer.is_time(dae_t))

    @property
    def class_name(self):
        """
        Return the class name
        """
        return self.__class__.__name__

    def _all_vars(self):
        """
        An OrderedDict of States, ExtStates, Algebs, ExtAlgebs
        """
        return OrderedDict(list(self.states.items()) +
                           list(self.states_ext.items()) +
                           list(self.algebs.items()) +
                           list(self.algebs_ext.items())
                           )

    def _iter_vars(self):
        """
        Variables to be iteratively initialized
        """
        all_vars = OrderedDict(self.cache.all_vars)
        for name, instance in self.cache.all_vars.items():
            if not instance.v_iter:
                all_vars.pop(name)
        return all_vars

    def _all_vars_names(self):
        out = []
        for instance in self.cache.all_vars.values():
            out += instance.get_names()
        return out

    def _all_params(self):
        # the service stuff should not be moved to variables.
        return OrderedDict(list(self.num_params.items()) +
                           list(self.services.items()) +
                           list(self.services_ext.items()) +
                           list(self.services_ops.items()) +
                           list(self.services_subs.items()) +
                           list(self.discrete.items())
                           )

    def _all_params_names(self):
        out = []
        for instance in self.cache.all_params.values():
            out += instance.get_names()
        return out

    def _algebs_and_ext(self):
        return OrderedDict(list(self.algebs.items()) +
                           list(self.algebs_ext.items()))

    def _states_and_ext(self):
        return OrderedDict(list(self.states.items()) +
                           list(self.states_ext.items()))

    def _services_and_ext(self):
        return OrderedDict(list(self.services.items()) +
                           list(self.services_ext.items()))

    def _vars_ext(self):
        return OrderedDict(list(self.states_ext.items()) +
                           list(self.algebs_ext.items()))

    def _vars_int(self):
        return OrderedDict(list(self.states.items()) +
                           list(self.algebs.items()))

    def _v_getters(self):
        out = OrderedDict()
        for name, var in self.cache.all_vars.items():
            if var.v_inplace:
                continue
            out[name] = var
        return out

    def _v_adders(self):
        out = OrderedDict()
        for name, var in self.cache.all_vars.items():
            if var.v_inplace is True:
                continue
            if var.v_str is None and var.v_iter is None:
                continue
            if var.v_setter is True:
                continue

            out[name] = var
        return out

    def _v_setters(self):
        out = OrderedDict()
        for name, var in self.cache.all_vars.items():
            if var.v_inplace is True:
                continue
            if var.v_str is None and var.v_iter is None:
                continue
            if var.v_setter is False:
                continue

            out[name] = var
        return out

    def _e_adders(self):
        out = OrderedDict()
        for name, var in self.cache.all_vars.items():
            if var.e_inplace is True:
                continue
            if var.e_str is None:
                continue
            if var.e_setter is True:
                continue

            out[name] = var
        return out

    def _e_setters(self):
        out = OrderedDict()
        for name, var in self.cache.all_vars.items():
            if var.e_inplace is True:
                continue
            if var.e_str is None:
                continue
            if var.e_setter is False:
                continue

            out[name] = var
        return out

    def _input_vars(self):
        out = list()
        for name, var in self.cache.all_vars.items():
            if var.is_input:
                out.append(name)
        return out

    def _output_vars(self):
        out = list()
        for name, var in self.cache.all_vars.items():
            if var.is_output:
                out.append(name)
        return out

    def set_in_use(self):
        """
        Set the `in_use` attribute. Called at the end of ``System.collect_ref``.

        This function is overloaded by models with `BackRef` to disable calls when no model is referencing.
        Models with no back references will have internal variable addresses assigned but external addresses
        being empty.

        For internal equations that have external variables, the row indices will be non-zeros, while the col
        indices will be empty, which causes an error when updating Jacobians.

        Setting `self.in_use` to False when `len(back_ref_instance.v) == 0` avoids this error. See COI.
        """
        self.in_use = True

    def list2array(self):
        """
        Convert all the value attributes ``v`` to NumPy arrays.

        Value attribute arrays should remain in the same address afterwards.
        Namely, all assignments to value array should be operated in place (e.g., with [:]).
        """

        for instance in self.num_params.values():
            instance.to_array()

        for instance in self.cache.services_and_ext.values():
            instance.assign_memory(self.n)

        for instance in self.discrete.values():
            instance.list2array(self.n)

    def a_reset(self):
        """
        Reset addresses to empty and reset flags.address to ``False``.
        """
        for var in self.cache.all_vars.values():
            var.reset()
        self.flags.address = False
        self.flags.initialized = False

    def e_clear(self):
        """
        Clear equation value arrays associated with all internal variables.
        """
        for instance in self.cache.all_vars.values():
            if instance.e_inplace:
                continue
            instance.e[:] = 0

    def v_numeric(self, **kwargs):
        """
        Custom variable initialization function.
        """
        pass

    def g_numeric(self, **kwargs):
        """
        Custom gcall functions. Modify equations directly.
        """
        pass

    def f_numeric(self, **kwargs):
        """
        Custom fcall functions. Modify equations directly.
        """
        pass

    def s_numeric(self, **kwargs):
        """
        Custom service value functions. Modify ``Service.v`` directly.
        """
        pass

    def s_numeric_var(self, **kwargs):
        """
        Custom variable service value functions. Modify ``VarService.v`` directly.

        This custom numerical function is evaluated at each step/iteration before equation update.
        """
        pass

    def j_numeric(self, **kwargs):
        """
        Custom numeric update functions.

        This function should append indices to `_ifx`, `_jfx`, and append anonymous functions to `_vfx`.
        It is only called once by `store_sparse_pattern`.
        """
        pass

    def doc(self, max_width=78, export='plain'):
        """
        Retrieve model documentation as a string.
        """
        return self.docum.get(max_width=max_width, export=export)

    def prepare(self, quick=False, pycode_path=None, yapf_pycode=False):
        """
        Symbolic processing and code generation.
        """
        logger.debug(f"Generating code for {self.class_name}")
        self.calls.md5 = self.get_md5()

        self.syms.generate_symbols()
        self.syms.generate_subs_expr()
        self.syms.generate_equations()
        self.syms.generate_services()
        self.syms.generate_jacobians()
        self.syms.generate_init()

        self.syms.generate_pycode(pycode_path=pycode_path,
                                  yapf_pycode=yapf_pycode,
                                  )
        if quick is False:
            self.syms.generate_pretty_print()

    def get_md5(self):
        """
        Return the md5 hash of concatenated equation strings.
        """
        import hashlib
        md5 = hashlib.md5()

        for name in self.cache.all_params.keys():
            md5.update(str(name).encode())

        for name in self.config.as_dict().keys():
            md5.update(str(name).encode())

        for name, item in self.cache.all_vars.items():
            md5.update(str(name).encode())

            if item.v_str is not None:
                md5.update(str(item.v_str).encode())
            if item.v_iter is not None:
                md5.update(str(item.v_iter).encode())
            if item.e_str is not None:
                md5.update(str(item.e_str).encode())
            if item.diag_eps is not None:
                md5.update(str(item.diag_eps).encode())

        for name, item in self.services.items():
            md5.update(str(name).encode())

            if item.v_str is not None:
                md5.update(str(item.v_str).encode())

            md5.update(str(int(item.sequential)).encode())

        for name, item in self.discrete.items():
            md5.update(str(name).encode())
            md5.update(str(','.join(item.export_flags)).encode())

        return md5.hexdigest()

    def post_init_check(self):
        """
        Post init checking. Warns if values of `InitChecker` is not True.
        """
        self.get_inputs(refresh=True)

        if self.system.config.warn_abnormal:
            for item in self.services_icheck.values():
                item.check()

    def numba_jitify(self, parallel=False, cache=True, nopython=False):
        """
        Convert equation residual calls, Jacobian calls,
        and variable service calls into JIT compiled functions.

        This function can be turned on by setting
        ``System.config.numba`` to ``1``.

        Warnings
        --------
        This feature is experimental and does not guarantee a speed up.
        In fact, the program will likely end up being slower due to compilation.
        """
        if self.system.config.numba != 1:
            return

        if self.flags.jited is True:
            return

        kwargs = {'parallel': parallel,
                  'cache': cache,
                  'nopython': nopython,
                  }

        self.calls.f = to_jit(self.calls.f, **kwargs)
        self.calls.g = to_jit(self.calls.g, **kwargs)
        self.calls.sns = to_jit(self.calls.sns, **kwargs)

        for jname in self.calls.j:
            self.calls.j[jname] = to_jit(self.calls.j[jname], **kwargs)

        for name, instance in self.services_var_seq.items():
            if instance.v_str is not None:
                self.calls.s[name] = to_jit(self.calls.s[name], **kwargs)

        self.flags.jited = True

    def precompile(self):
        """
        Trigger numba compilation for this model.

        This function requires the system to be setup, i.e.,
        memory allocated for storage.
        """

        self.get_inputs()
        if self.n == 0:
            self.mock_refresh_inputs()
        self.refresh_inputs_arg()

        if callable(self.calls.f):
            self.calls.f(*self.f_args)

        if callable(self.calls.g):
            self.calls.g(*self.g_args)

        if callable(self.calls.sns):
            self.calls.sns(*self.sns_args)

        for jname, jfunc in self.calls.j.items():
            jfunc(*self.j_args[jname])

        for name, instance in self.services_var_seq.items():
            if instance.v_str is not None:
                self.calls.s[name](*self.s_args[name])

    def __repr__(self):
        dev_text = 'device' if self.n == 1 else 'devices'

        return f'{self.class_name} ({self.n} {dev_text}) at {hex(id(self))}'

    def init(self, routine):
        """
        Numerical initialization of a model.

        Initialization sequence:
        1. Sequential initialization based on the order of definition
        2. Use Newton-Krylov method for iterative initialization
        3. Custom init
        """

        # evaluate `ConstService` and `VarService`
        self.s_update()
        self.s_update_var()

        # find out if variables need to be initialized for `routine`
        flag_name = routine + '_init'

        if not hasattr(self.flags, flag_name) or getattr(self.flags, flag_name) is None:
            do_init = getattr(self.flags, routine)
        else:
            do_init = getattr(self.flags, flag_name)

        sys_debug = self.system.options.get("init")

        logger.debug('========== %s has <%s> = %s ==========',
                     self.class_name, flag_name, do_init)

        if do_init:
            kwargs = self.get_inputs(refresh=True)

            logger.debug('Initialization sequence:')
            seq_str = ' -> '.join([str(i) for i in self.calls.init_seq])
            logger.debug('\n'.join(wrap(seq_str, 70)))
            logger.debug("%s: assignment initialization phase begins", self.class_name)

            for idx, name in enumerate(self.calls.init_seq):
                debug_flag = sys_debug or (name in self.debug_equations)

                # single variable - do assignment
                if isinstance(name, str):
                    _log_init_debug(debug_flag,
                                    "%s: entering <%s> assignment init",
                                    self.class_name, name)

                    instance = self.__dict__[name]
                    if instance.discrete is not None:
                        _log_init_debug(debug_flag, "%s: evaluate discrete <%s>", name, instance.discrete)
                        _eval_discrete(instance, self.config.allow_adjust,
                                       self.config.adjust_lower, self.config.adjust_upper)

                    if instance.v_str is not None:
                        arg_print = OrderedDict()
                        if debug_flag:
                            for a, b in zip(self.calls.ia_args[name], self.ia_args[name]):
                                arg_print[a] = b

                        if not instance.v_str_add:
                            # assignment is for most variable initialization
                            _log_init_debug(debug_flag, "%s: new values will be assigned (=)", name)
                            instance.v[:] = self.calls.ia[name](*self.ia_args[name])

                        else:
                            # in-place add initial values.
                            # Voltage compensators can set part of the `v` of exciters.
                            # Exciters will then set the bus voltage part.
                            _log_init_debug(debug_flag, "%s: new values will be in-place added (+=)", name)
                            instance.v[:] += self.calls.ia[name](*self.ia_args[name])

                        arg_print[name] = instance.v

                        if debug_flag:
                            for key, val in arg_print.items():
                                if isinstance(val, (int, float, np.floating, np.integer)) or \
                                        isinstance(val, np.ndarray) and val.ndim == 0:

                                    arg_print[key] = val * np.ones_like(instance.v)
                                if isinstance(val, np.ndarray) and val.dtype == complex:
                                    arg_print[key] = [str(i) for i in val]

                            tab = Tab(title="v_str of %s is '%s'" % (name, instance.v_str),
                                      header=["idx", *self.calls.ia_args[name], name],
                                      data=list(zip(self.idx.v, *arg_print.values())),
                                      )
                            _log_init_debug(debug_flag, tab.draw())

                    # single variable iterative solution
                    if name in self.calls.ii:
                        _log_init_debug(debug_flag,
                                        "\n%s: entering <%s> iterative init",
                                        self.class_name,
                                        pprint.pformat(name))

                        self.solve_iter(name, kwargs)

                        _log_init_debug(debug_flag,
                                        "%s new values are %s", name, self.__dict__[name].v)

                # multiple variables, iterative
                else:
                    _log_init_debug(debug_flag, "\n%s: entering <%s> iterative init",
                                    self.class_name, name)

                    for vv in name:
                        instance = self.__dict__[vv]

                        if instance.discrete is not None:
                            _log_init_debug(debug_flag, "%s: evaluate discrete <%s>",
                                            name, instance.discrete)

                            _eval_discrete(instance, self.config.allow_adjust,
                                           self.config.adjust_lower, self.config.adjust_upper)
                        if instance.v_str is not None:
                            _log_init_debug(debug_flag, "%s: v_str = %s", vv, instance.v_str)

                            arg_print = OrderedDict()
                            if debug_flag:
                                for a, b in zip(self.calls.ia_args[vv], self.ia_args[vv]):
                                    arg_print[a] = b
                                _log_init_debug(debug_flag, pprint.pformat(arg_print))

                            instance.v[:] = self.calls.ia[vv](*self.ia_args[vv])

                    _log_init_debug(debug_flag, "\n%s: entering <%s> iterative init", self.class_name,
                                    pprint.pformat(name))
                    self.solve_iter(name, kwargs)

                    for vv in name:
                        instance = self.__dict__[vv]
                        _log_init_debug(debug_flag, "%s new values are %s", vv, instance.v)

            # call custom variable initializer after generated init
            kwargs = self.get_inputs(refresh=True)
            self.v_numeric(**kwargs)

        # call post initialization checking
        self.post_init_check()

        self.flags.initialized = True

    def solve_iter(self, name, kwargs):
        """
        Solve iterative initialization.
        """
        for pos in range(self.n):
            logger.debug("%s: iterative init for %s, device pos = %s",
                         self.class_name, name, pos)

            self.solve_iter_single(name, kwargs, pos)

    def solve_iter_single(self, name, inputs, pos):
        """
        Solve iterative initialization for one given device.
        """

        if isinstance(name, str):
            x0 = inputs[name][pos]
            name_concat = name
        else:
            x0 = np.ravel([inputs[item][pos] for item in name])
            name_concat = '_'.join(name)

        rhs = self.calls.ii[name_concat]
        jac = self.calls.ij[name_concat]
        ii_args = self.ii_args[name_concat]
        ij_args = self.ij_args[name_concat]

        # iteration setup
        maxiter = self.system.TDS.config.max_iter
        eps = self.system.TDS.config.tol
        niter = 0
        solved = False

        while niter < maxiter:
            logger.debug("iteration %s:", niter)

            i_args = [item[pos] for item in ii_args]  # all variables of one device at a time
            j_args = [item[pos] for item in ij_args]

            b = np.ravel(rhs(*i_args))
            A = jac(*j_args)

            logger.debug("A:\n%s", A)
            logger.debug("b:\n%s", b)

            mis = np.max(np.abs(b))
            if mis <= eps:
                solved = True
                break

            inc = - sp.linalg.lu_solve(sp.linalg.lu_factor(A), b)
            if np.isnan(inc).any():
                if self.u.v[pos] != 0:
                    logger.debug("%s: nan ignored in iterations for offline device pos = %s",
                                 self.class_name, pos)
                else:
                    logger.error("%s: nan error detected in iterations for device pos = %s",
                                 self.class_name, pos)
                break

            x0 += inc
            logger.debug("solved x0:\n%s", x0)

            for idx, item in enumerate(name):
                inputs[item][pos] = x0[idx]

            niter += 1

        if solved:
            for idx, item in enumerate(name):
                inputs[item][pos] = x0[idx]
        else:
            logger.warning(f"{self.class_name}: iterative initialization failed for {self.idx.v[pos]}.")

    def externalize(self):
        """
        Externalize internal data as a snapshot.
        """
        pass

    def internalize(self):
        """
        Internalize snapshot data.
        """
        pass

    def register_debug_equation(self, var_name):
        """
        Helper function to register a variable for debugging the initialization.

        This function needs to be called before calling ``TDS.init()``, and
        logging level needs to be set to ``DEBUG``.
        """

        self.debug_equations.append(var_name)


def _eval_discrete(instance, allow_adjust=True,
                   adjust_lower=False, adjust_upper=False):
    """
    Evaluate discrete components associated with a variable instance.
    Calls ``check_var()`` on the discrete components.

    For variables associated with limiters, limiters need to be evaluated
    before variable initialization.
    However, if any limit is hit, initialization is likely to fail.

    Parameters
    ----------
    instance : BaseVar
        instance of a variable
    allow_adjust : bool, optional
        True to enable overall adjustment
    adjust_lower : bool, optional
        True to adjust lower limits to the input values
    adjust_upper : bool, optional
        True to adjust upper limits to the input values

    """

    if not isinstance(instance.discrete, (list, tuple, set)):
        dlist = (instance.discrete,)
    else:
        dlist = instance.discrete
    for d in dlist:
        d.check_var(allow_adjust=allow_adjust,
                    adjust_lower=adjust_lower,
                    adjust_upper=adjust_upper,
                    is_init=True,
                    )
        d.check_eq(allow_adjust=allow_adjust,
                   adjust_lower=adjust_lower,
                   adjust_upper=adjust_upper,
                   is_init=True,
                   )


def to_jit(func: Union[Callable, None],
           parallel: bool = False,
           cache: bool = False,
           nopython: bool = False,
           ):
    """
    Helper function for converting a function to a numba jit-compiled function.

    Note that this function will be compiled just-in-time when first called,
    based on the argument types.
    """

    if func is not None:
        return numba.jit(func,
                         parallel=parallel,
                         cache=cache,
                         nopython=nopython,
                         )

    return func


def _log_init_debug(debug_flag, *args):
    """
    Helper function to log initialization debug message.
    """

    if debug_flag is True:
        logger.debug(*args)



from andes.linsolvers.cupy import CuPySolver
from andes.linsolvers.scipy import SpSolve
from andes.linsolvers.suitesparse import UMFPACKSolver, KLUSolver


class Solver:
    """
    Sparse matrix solver class.

    This class wraps UMFPACK, KLU, SciPy and CuPy solvers to provide an unified
    interface for solving sparse linear equations ``Ax = b``.

    Provides methods ``solve``, ``linsolve`` and ``clear``.
    """

    def __init__(self, sparselib='umfpack'):

        # solvers
        self.umfpack = UMFPACKSolver()
        self.klu = KLUSolver()
        self.spsolve = SpSolve()
        self.cupy = CuPySolver()

        # KLU as failsafe
        if sparselib not in self.__dict__:
            self.sparselib = 'klu'
        else:
            self.sparselib = sparselib

        self.worker = self.__dict__[self.sparselib]

    def solve(self, A, b):
        """
        Solve linear equations and cache factorizations if possible.

        Parameters
        ----------
        A : kvxopt.spmatrix
            Sparse N-by-N matrix
        b : kvxopt.matrix or numpy.ndarray
            Dense N-by-1 matrix

        Returns
        -------
        numpy.ndarray
            Dense N-by-1 array
        """
        return self.worker.solve(A, b)

    def linsolve(self, A, b):
        """
        Solve linear equations without caching facorization. Performs full factorization each call.

        Parameters
        ----------
        A : kvxopt.spmatrix
            Sparse N-by-N matrix
        b : kvxopt.matrix or numpy.ndarray
            Dense N-by-1 matrix

        Returns
        -------
        numpy.ndarray
            Dense N-by-1 array
        """
        return self.worker.linsolve(A, b)

    def clear(self):
        """
        Remove all cached objects.
        """
        self.worker.clear()



import os
import unittest

import andes
from andes.utils.paths import list_cases


class TestPaths(unittest.TestCase):
    def setUp(self) -> None:
        self.kundur = 'kundur/'
        self.matpower = 'matpower/'
        self.ieee14 = andes.get_case("ieee14/ieee14.raw")

    def test_tree(self):
        list_cases(self.kundur, no_print=True)
        list_cases(self.matpower, no_print=True)

    def test_addfile_path(self):
        path, case = os.path.split(self.ieee14)
        ss = andes.load('ieee14.raw', addfile='ieee14.dyr',
                        input_path=path, default_config=True,
                        )
        self.assertNotEqual(ss, None)

        ss = andes.run('ieee14.raw', addfile='ieee14.dyr',
                       input_path=path,
                       no_output=True, default_config=True,
                       )
        self.assertNotEqual(ss, None)

    def test_relative_path(self):
        ss = andes.run('ieee14.raw',
                       input_path=andes.get_case('ieee14/', check=False),
                       no_output=True, default_config=True,
                       )
        self.assertNotEqual(ss, None)

    def test_pert_file(self):
        """Test path of pert file"""
        path, case = os.path.split(self.ieee14)

        # --- with pert file ---
        ss = andes.run('ieee14.raw', pert='pert.py',
                       input_path=path, no_output=True, default_config=True,
                       )
        ss.TDS.init()
        self.assertIsNotNone(ss.TDS.callpert)

        # --- without pert file ---
        ss = andes.run('ieee14.raw',
                       input_path=path, no_output=True, default_config=True,
                       )
        ss.TDS.init()
        self.assertIsNone(ss.TDS.callpert)



"""
Module with extra SymPy classes.
"""

import sympy


class Indicator(sympy.Expr):
    """
    Indicator class for printing SymPy Relational.

    Relational expressions in SymPy need to be wrapped by `Indicator`.

    Examples
    --------
    To compare ``dae_t`` with ``0``, one need to use
    ``Indicator(dae_t < 0)```.
    """

    def _numpycode(self, printer):
        return printer._print(self.args[0])


class FixPiecewise(sympy.Piecewise):
    """
    A derived Piecewise that fixes the printing of ``select`` to allow compilation with numba.

    See: https://github.com/sympy/sympy/issues/15014
    """

    def _numpycode(self, printer):
        """
        Updated numpy code printer.
        """

        def broadcastarg(arg):
            if arg.has(sympy.Symbol):
                return printer._print(arg)
            if arg == 0:
                return '__zeros'

            return printer._print(arg * sympy.Symbol('__ones'))

        def broadcastcond(cond):
            if cond.has(sympy.Symbol):
                return printer._print(cond)

            if bool(cond):
                return '__trues'
            else:
                return '__falses'

        # Piecewise function printer

        exprs = '[{}]'.format(','.join(broadcastarg(arg.expr) for arg in self.args))
        conds = '[{}]'.format(','.join(broadcastcond(arg.cond) for arg in self.args))
        # If [default_value, True] is a (expr, cond) sequence in a Piecewise object
        #     it will behave the same as passing the 'default' kwarg to select()
        #     *as long as* it is the last element in self.args.
        # If this is not the case, it may be triggered prematurely.
        return '{}({}, {}, default={})'.format(
            printer._module_format('select'), conds, exprs,
            printer._print(sympy.S.NaN))



import os
import unittest

import numpy as np

import andes
from andes.utils.paths import get_case


class Test5Bus(unittest.TestCase):
    def setUp(self) -> None:
        self.ss = andes.main.load(get_case('5bus/pjm5bus.json'),
                                  default_config=True,
                                  no_output=True,
                                  )

    def test_names(self):
        self.assertTrue('Bus' in self.ss.models)
        self.assertTrue('PQ' in self.ss.models)

    def test_count(self):
        self.assertEqual(self.ss.Bus.n, 5)
        self.assertEqual(self.ss.PQ.n, 3)
        self.assertEqual(self.ss.PV.n, 3)
        self.assertEqual(self.ss.Slack.n, 1)
        self.assertEqual(self.ss.Line.n, 7)
        self.assertEqual(self.ss.GENCLS.n, 4)
        self.assertEqual(self.ss.TG2.n, 4)

    def test_idx(self):
        self.assertSequenceEqual(self.ss.Bus.idx.v, [0, 1, 2, 3, 4])
        self.assertSequenceEqual(self.ss.Area.idx.v, [1, 2, 3])

    def test_cache_refresh(self):
        self.ss.Bus.cache.refresh()

    def test_as_df(self):
        self.ss.Bus.as_df()
        self.ss.Bus.as_df(vin=True)

    def test_init_order(self):
        self.ss.Bus.get_init_order()

    def test_pflow_reset(self):
        self.ss.PFlow.run()
        if self.ss.PFlow.config.init_tds == 0:
            self.ss.reset()
            self.ss.PFlow.run()

    def test_alter_param(self):
        self.ss.PV.alter('v0', 2, 0.98)
        self.assertEqual(self.ss.PV.v0.v[1], 0.98)
        self.ss.PFlow.run()


class TestKundur2AreaEIG(unittest.TestCase):
    """
    Test Kundur's 2-area system
    """

    def test_xlsx_eig_run(self):
        self.xlsx = get_case("kundur/kundur_full.xlsx")
        self.ss = andes.run(self.xlsx,
                            default_config=True,
                            no_output=True,
                            )

        self.ss.EIG.run()


class TestKundur2AreaPSSE(unittest.TestCase):
    """
    Test Kundur's 2-area system in PSS/E format
    """

    def setUp(self) -> None:
        raw = get_case("kundur/kundur.raw")
        dyr = get_case("kundur/kundur_full.dyr")
        self.ss_psse = andes.run(raw, addfile=dyr, default_config=True,
                                 no_output=True,
                                 )

    def test_psse_tds_run_with_stats(self):
        self.ss_psse.config.save_stats = 1
        self.ss_psse.TDS.config.tf = 3
        self.ss_psse.TDS.run()

        self.assertEqual(self.ss_psse.exit_code, 0, "Exit code is not 0.")

    def test_psse_eig_run(self):
        self.ss_psse.EIG.run()

        self.assertEqual(self.ss_psse.exit_code, 0, "Exit code is not 0.")

    def test_kundur_psse2xlsx(self):
        output_name = 'test_kundur_convert.xlsx'
        andes.io.dump(self.ss_psse, 'xlsx', full_path=output_name)
        os.remove(output_name)


class TestNPCCRAW(unittest.TestCase):
    """
    Test NPCC system in the RAW format.
    """

    def test_npcc_raw(self):
        self.ss = andes.run(get_case('npcc/npcc.raw'),
                            default_config=True,
                            no_output=True,
                            )

    def test_npcc_raw_tds(self):
        self.ss = andes.run(get_case('npcc/npcc.raw'),
                            verbose=50,
                            routine='tds',
                            no_output=True,
                            profile=True,
                            tf=10,
                            default_config=True,
                            )

        self.ss.dae.print_array('f')
        self.ss.dae.print_array('g')
        self.ss.dae.print_array('f', tol=1e-4)
        self.ss.dae.print_array('g', tol=1e-4)

    def test_npcc_raw_convert(self):
        self.ss = andes.run(get_case('npcc/npcc.raw'),
                            convert=True,
                            default_config=True,
                            )

        os.remove(self.ss.files.dump)
        self.assertEqual(self.ss.exit_code, 0, "Exit code is not 0.")

    def test_npcc_raw2json_convert(self):
        self.ss = andes.run(get_case('npcc/npcc.raw'),
                            convert='json',
                            default_config=True,
                            )

        self.ss2 = andes.run('npcc.json',
                             default_config=True,
                             no_output=True,
                             )

        os.remove(self.ss.files.dump)
        self.assertEqual(self.ss2.exit_code, 0, "Exit code is not 0.")

    def test_read_json_from_memory(self):
        fd = open(get_case('ieee14/ieee14_zip.json'), 'r')

        ss = andes.main.System(default_config=True,
                               no_output=True,
                               )
        andes.io.json.read(ss, fd)
        ss.setup()
        ss.PFlow.run()

        fd.close()
        self.assertEqual(ss.exit_code, 0, "Exit code is not 0.")

    def test_read_mpc_from_memory(self):
        fd = open(get_case('matpower/case14.m'), 'r')

        ss = andes.main.System(default_config=True,
                               no_output=True,
                               )
        andes.io.matpower.read(ss, fd)
        ss.setup()
        ss.PFlow.run()

        fd.close()
        self.assertEqual(ss.exit_code, 0, "Exit code is not 0.")

    def test_read_psse_from_memory(self):
        fd_raw = open(get_case('npcc/npcc.raw'), 'r')
        fd_dyr = open(get_case('npcc/npcc_full.dyr'), 'r')

        ss = andes.main.System(default_config=True,
                               no_output=True,
                               )
        # suppress out-of-normal info
        ss.config.warn_limits = 0
        ss.config.warn_abnormal = 0

        andes.io.psse.read(ss, fd_raw)
        andes.io.psse.read_add(ss, fd_dyr)
        ss.setup()
        ss.PFlow.run()
        ss.TDS.init()

        fd_raw.close()
        fd_dyr.close()
        self.assertEqual(ss.exit_code, 0, "Exit code is not 0.")


class TestPlot(unittest.TestCase):
    def test_kundur_plot(self):
        ss = andes.run(get_case('kundur/kundur_full.json'),
                       routine='tds',
                       tf=2.0,
                       no_output=True,
                       default_config=True,
                       )

        ss.TDS.load_plotter()

        ss.TDS.plt.plot(ss.Bus.v, ylabel="Bus Voltages [pu]",
                        title='Bus Voltage Plot',
                        left=0.2, right=1.5,
                        ymin=0.95, ymax=1.05, legend=True, grid=True, greyscale=True,
                        hline=[1.01, 1.02], vline=[0.5, 0.8],
                        dpi=80, line_width=1.2, font_size=11, show=False,
                        )

        self.assertEqual(ss.exit_code, 0, "Exit code is not 0.")


class TestCOI(unittest.TestCase):
    def test_kundur_COI(self):
        ss = get_case('kundur/kundur_coi.json')
        exit_code = andes.run(ss,
                              routine='tds',
                              no_output=True,
                              tf=0.1,
                              cli=True,
                              default_config=True,
                              )

        self.assertEqual(exit_code, 0, "Exit code is not 0.")

    def test_kundur_COI_empty(self):
        ss = get_case('kundur/kundur_coi_empty.json')

        exit_code = andes.run(ss,
                              routine='tds',
                              no_output=True,
                              tf=0.1,
                              cli=True,
                              default_config=True,
                              )

        self.assertEqual(exit_code, 0, "Exit code is not 0.")


class TestVSC(unittest.TestCase):
    """Test case =for VSC power flow model"""

    def test_kundur_vsc(self):
        """Test power flow exit code"""

        ss = get_case('kundur/kundur_vsc.json')
        exit_code = andes.run(ss,
                              routine='pflow',
                              no_output=True,
                              cli=True,
                              default_config=True,
                              )

        self.assertEqual(exit_code, 0, "Exit code is not 0.")


class TestShuntSw(unittest.TestCase):
    """Test class for switched shunt."""

    def test_shuntsw(self):
        """
        Test `ShuntSw` class.
        """

        case = get_case('ieee14/ieee14_shuntsw.json')
        ss = andes.run(case,
                       no_output=True,
                       default_config=True,
                       )

        self.assertEqual(ss.exit_code, 0, "Exit code is not 0.")

        np.testing.assert_almost_equal(ss.ShuntSw.beff.v, [0.1, 0.1])
        np.testing.assert_almost_equal(ss.ShuntSw.beff.bcs[0],
                                       [0., 0.025, 0.05, 0.075, 0.1, 0.125])
        np.testing.assert_almost_equal(ss.ShuntSw.beff.bcs[1],
                                       [0., 0.05, 0.1, 0.15, 0.2, 0.25])


class TestIslands(unittest.TestCase):
    """Test power flow with two islands"""

    def test_islands(self):
        ss = andes.run(get_case('kundur/kundur_islands.json'),
                       no_output=True, default_config=True)

        self.assertEqual(ss.exit_code, 0, "Exit code is not 0.")
        self.assertEqual(len(ss.Bus.islands), 2)


class TestCaseInit(unittest.TestCase):
    """
    Test if initializations pass.
    """

    def test_pvd1_init(self):
        """
        Test if PVD1 model initialization works.
        """
        ss = andes.run(get_case('ieee14/ieee14_pvd1.json'),
                       no_output=True,
                       default_config=True,
                       )
        ss.config.warn_limits = 0
        ss.config.warn_abnormal = 0

        ss.TDS.init()

        self.assertEqual(ss.exit_code, 0, "Exit code is not 0.")

    def test_exac1_init(self):
        """
        Test EXAC1 initialization with one TGOV1 at lower limit.
        """
        ss = andes.load(get_case('ieee14/ieee14_exac1.json'),
                        no_output=True,
                        default_config=True,
                        )
        ss.PV.config.pv2pq = 1
        ss.PFlow.run()

        # suppress EXAC1 warning from select
        np.seterr(invalid='ignore')

        ss.config.warn_limits = 0
        ss.config.warn_abnormal = 0

        ss.TDS.init()

        self.assertEqual(ss.exit_code, 0, "Exit code is not 0.")



"""
Output TXT file formatter.
"""

import os
from typing import Iterable

from andes.shared import np


def dump_data(text, header, rowname, data, file, width=18, precision=5):
    out = ''

    os.makedirs(os.path.abspath(os.path.dirname(file)), exist_ok=True)
    with open(file, 'w') as fid:

        for Text, Header, Rowname, Data in zip(text, header, rowname, data):
            # Write Text
            if Text:
                fid.writelines(Text)

            # determine the width for the first column (usually names)
            width_first = width
            if isinstance(Rowname, Iterable) and len(Rowname) > 0:
                for item in Rowname:
                    if isinstance(item, Iterable) and len(item) > width_first:
                        width_first = len(item)

            # Write Header
            if Header:
                ncol = len(Header)
                s = ' ' * width_first
                s += '{:>{width}s}' * ncol + '\n'
                fid.writelines(s.format(*Header, width=width))
                fid.write('\n')

            # Append Rowname to Data
            # Data is a list of column lists
            if Rowname is not None:
                ncol = 0
                for idx, item in enumerate(Rowname):  # write by row as always
                    if Data is None:
                        out = ''
                    elif isinstance(Data, (int, float, str)):
                        out = [Data]
                    elif isinstance(Data, (list, tuple, np.ndarray)):
                        if isinstance(Data[0], (int, float)):  # is a list of numbers
                            out = [Data[idx]]
                        elif isinstance(Data[0], (list, np.ndarray)):  # list of list in Data
                            ncol = len(Data)
                            out = [Data[i][idx] for i in range(ncol)]
                    else:
                        print(Data)
                        print('Unexpected Data during output, in formats/txt.py')

                    s = '{:{width_first}s}'  # for row header
                    for ii, col in enumerate(out):
                        if isinstance(col, (int, float)):
                            s += '{:>{width}.{precision}g}'
                        elif isinstance(col, str):
                            if len(col) > width:
                                out[ii] = col[:width]
                            s += '{:>{width}s}'
                        elif col is None:
                            out[ii] = 'None'
                            s += '{:>{width}s}'
                        else:
                            pass
                    s += '\n'

                    fid.write(
                        s.format(
                            str(item), *out, width_first=width_first, width=width, precision=precision))
            fid.write('\n')



import unittest

import andes


class TestTDSResume(unittest.TestCase):
    def test_tds_resume(self):
        case = andes.get_case('kundur/kundur_full.json')

        ss = andes.run(case, routine='tds', tf=0.1, no_output=True,
                       default_config=True)

        ss.TDS.config.tf = 0.5
        ss.TDS.run()

        self.assertEqual(ss.exit_code, 0.0)

        # check if time stamps are correctly stored
        self.assertIn(0.1, ss.dae.ts._ys)
        self.assertIn(0.5, ss.dae.ts._ys)

        # check if values are correctly stored
        self.assertTrue(len(ss.dae.ts._ys[0.1]) > 0)
        self.assertTrue(len(ss.dae.ts._ys[0.5]) > 0)

        # check if `get_data` works
        data1 = ss.dae.ts.get_data((ss.GENROU.omega, ss.GENROU.v))
        data2 = ss.dae.ts.get_data((ss.GENROU.omega, ss.GENROU.v), a=[2])
        nt = len(ss.dae.ts.t)

        self.assertEqual(data1.shape[1], 8)
        self.assertEqual(data1.shape[0], nt)

        self.assertEqual(data2.shape[1], 2)
        self.assertEqual(data2.shape[0], nt)



"""
Symbolic processor class for ANDES models.
"""

import inspect
import logging
import os
import pprint
from collections import OrderedDict, defaultdict

import numpy as np
import sympy as sp
from andes.shared import dilled_vars
from andes.thirdparty.npfunc import safe_div
from andes.thirdparty.sympymod import FixPiecewise
from andes.utils.paths import get_pycode_path

logger = logging.getLogger(__name__)

select_args_add = ["__zeros", "__ones", "__falses", "__trues"]


# the line below caches Piecewise instances
sp.OldPiecewise = sp.Piecewise

sp.Piecewise = FixPiecewise


class SymProcessor:
    """
    A helper class for symbolic processing and code generation.

    Parameters
    ----------
    parent : Model
        The `Model` instance to process

    Attributes
    ----------
    xy : sympy.Matrix
        variables pretty print in the order of State, ExtState, Algeb, ExtAlgeb
    f : sympy.Matrix
        differential equations pretty print
    g : sympy.Matrix
        algebraic equations pretty print
    df : sympy.SparseMatrix
        df /d (xy) pretty print
    dg : sympy.SparseMatrix
        dg /d (xy) pretty print
    inputs_dict : OrderedDict
        All possible symbols in equations, including variables, parameters, discrete flags, and
        config flags. It has the same variables as what ``get_inputs()`` returns.
    vars_dict : OrderedDict
        variable-only symbols, which are useful when getting the Jacobian matrices.

    """

    def __init__(self, parent):

        self.parent = parent
        # symbols that are input to lambda functions
        # including parameters, variables, services, configs, and scalars (dae_t, sys_f, sys_mva)
        self.inputs_dict = OrderedDict()
        self.lambdify_func = [dict(), 'numpy']

        self.vars_dict = OrderedDict()
        self.vars_int_dict = OrderedDict()   # internal variables only
        self.vars_list = list()
        self.substitution_map = {}    # mapping of ``SubsService`` to its expression

        self.f_list, self.g_list = list(), list()  # symbolic equations in lists
        self.f_matrix, self.g_matrix, self.s_matrix = list(), list(), list()  # equations in matrices
        self.s_syms = list()

        # pretty print of variables
        self.xy = list()  # variables in the order of states, algebs
        self.f, self.g, self.s = list(), list(), list()
        self.df, self.dg = None, None

        # get references to the parent attributes
        self.calls = parent.calls
        self.cache = parent.cache
        self.config = parent.config
        self.class_name = parent.class_name
        self.tex_names = OrderedDict()

    def generate_symbols(self):
        """
        Generate symbols for symbolic equation generations.

        This function should run before other generate equations.

        Attributes
        ----------
        inputs_dict : OrderedDict
            name-symbol pair of all parameters, variables and configs

        vars_dict : OrderedDict
            name-symbol pair of all variables, in the order of (states_and_ext + algebs_and_ext)

        """

        logger.debug('- Generating symbols for %s', self.class_name)

        # clear symbols storage
        self.f_list, self.g_list = list(), list()
        self.f_matrix, self.g_matrix = sp.Matrix([]), sp.Matrix([])

        # process tex_names defined in model
        # -----------------------------------------------------------
        for key in self.parent.tex_names.keys():
            self.tex_names[key] = sp.Symbol(self.parent.tex_names[key])
        for instance in self.parent.discrete.values():
            for name, tex_name in zip(instance.get_names(), instance.get_tex_names()):
                self.tex_names[name] = tex_name
        # -----------------------------------------------------------

        for var in self.cache.all_params_names:
            self.inputs_dict[var] = sp.Symbol(var)

        for var in self.cache.all_vars_names:
            tmp = sp.Symbol(var)
            self.vars_dict[var] = tmp
            self.inputs_dict[var] = tmp
            if var in self.cache.vars_int:
                self.vars_int_dict[var] = tmp

        # store tex names defined in `self.config`
        for key in self.config.as_dict():
            tmp = sp.Symbol(key)
            self.inputs_dict[key] = tmp
            if key in self.config.tex_names:
                self.tex_names[tmp] = sp.Symbol(self.config.tex_names[key])

        # store tex names for pretty printing replacement later
        for var in self.inputs_dict:
            if var in self.parent.__dict__ and self.parent.__dict__[var].tex_name is not None:
                self.tex_names[sp.Symbol(var)] = sp.Symbol(self.parent.__dict__[var].tex_name)

        # additional variables by conventions
        self.inputs_dict['dae_t'] = sp.Symbol('dae_t')
        self.inputs_dict['sys_f'] = sp.Symbol('sys_f')
        self.inputs_dict['sys_mva'] = sp.Symbol('sys_mva')

        # custom functions
        self.lambdify_func[0]['Indicator'] = lambda x: x
        self.lambdify_func[0]['imag'] = np.imag
        self.lambdify_func[0]['real'] = np.real
        self.lambdify_func[0]['safe_div'] = safe_div

        self.vars_list = list(self.vars_dict.values())  # useful for ``.jacobian()``

    def _check_expr_symbols(self, expr):
        """
        Check if expression contains unknown symbols.
        """
        fs = expr.free_symbols
        for item in fs:
            if item not in self.inputs_dict.values():
                raise ValueError(f'{self.class_name} expression "{expr}" contains unknown symbol "{item}"')

        fs = sorted(fs, key=lambda s: s.name)
        return fs

    def generate_subs_expr(self):
        """
        Generate expressions for substituting ``SubsService``.
        """
        for name, instance in self.parent.services_subs.items():
            if instance.v_str is not None:
                expr = sp.sympify(instance.v_str, locals=self.inputs_dict)
                self.substitution_map[self.inputs_dict[name]] = expr

    def _do_substitute(self, input_expr):
        """
        Helper function to substitute ``SubsService`` with its expression.
        """
        return input_expr.subs(self.substitution_map)

    def generate_equations(self):
        """
        Generate equations.

        The pretty-print equations in matrices can be accessed in
        ``self.f_matrix`` and ``self.g_matrix``.

        """

        logger.debug('- Generating equations for %s', self.class_name)

        self.f_list, self.g_list = list(), list()

        self.calls.f = None
        self.calls.g = None
        self.calls.f_args = list()
        self.calls.g_args = list()

        vars_list = [self.cache.states_and_ext, self.cache.algebs_and_ext]
        expr_list = [self.f_list, self.g_list]

        eqn_names = ['f', 'g']
        eqn_args = [self.calls.f_args, self.calls.g_args]

        for vlist, elist, ename, eargs in zip(vars_list, expr_list, eqn_names, eqn_args):
            sym_args = list()
            for name, instance in vlist.items():
                if instance.e_str is None:
                    elist.append(0)
                else:
                    try:
                        expr = sp.sympify(instance.e_str, locals=self.inputs_dict)
                        expr = self._do_substitute(expr)
                    except (sp.SympifyError, TypeError) as e:
                        logger.error('Error parsing equation "%s "for %s.%s',
                                     instance.e_str, instance.owner.class_name, name)
                        raise e

                    free_syms = self._check_expr_symbols(expr)
                    for s in free_syms:
                        if s not in sym_args:
                            sym_args.append(s)
                            eargs.append(str(s))

                    elist.append(expr)

            sym_args = sorted(sym_args, key=lambda s: s.name)
            eargs.sort()

            if len(elist) == 0 or not any(elist):  # `any`, not `all`
                self.calls.__dict__[ename] = None
            else:
                self.calls.__dict__[ename] = sp.lambdify(sym_args, tuple(elist),
                                                         modules=self.lambdify_func)

                # manually append additional arguments for select.
                if 'select' in inspect.getsource(self.calls.__dict__[ename]):
                    eargs.extend(select_args_add)

        # convert to SymPy matrices
        self.f_matrix = sp.Matrix(self.f_list)
        self.g_matrix = sp.Matrix(self.g_list)

    def generate_services(self):
        """
        Generate calls for services, including ``ConstService`` and
        ``VarService`` among others. Sequence is preserved due to possible
        dependency.

        All ``VarService`` with ``sequential = False`` are generated into one
        function, ``calls.sns``.

        Starting from SymPy 1.9, ``Matrix`` no longer supports non-Expr objects.
        Due to the use of logical conditions in services, service expressions
        will not be converted to SymPy Matrix. The dictionary to look up service
        name and expression is in ``self.s_syms``.
        """

        s_syms = OrderedDict()
        s_args = OrderedDict()
        s_calls = OrderedDict()
        sns_args = list()   # ``sns`` means services that are non-sequential
        sns_calls = None

        s_calls_nonseq = list()
        s_calls_nonseq_args = list()

        for name, instance in self.parent.services.items():
            v_str = '0' if instance.v_str is None else instance.v_str
            try:
                expr = sp.sympify(v_str, locals=self.inputs_dict)
                expr = self._do_substitute(expr)
            except (sp.SympifyError, TypeError) as e:
                logger.error(f'Error parsing equation for {instance.owner.class_name}.{name}')
                raise e

            fs = self._check_expr_symbols(expr)
            s_syms[name] = expr
            args_expr = [str(i) for i in fs]

            if instance.sequential is True:
                s_args[name] = args_expr
                s_calls[name] = sp.lambdify(s_args[name], s_syms[name], modules=self.lambdify_func)
                if 'select' in inspect.getsource(s_calls[name]):
                    s_args[name].extend(select_args_add)
            else:
                s_calls_nonseq.append(expr)
                s_calls_nonseq_args.extend(args_expr)

        if len(s_calls_nonseq) > 0:
            sns_args = sorted(list(set(s_calls_nonseq_args)))
            sns_calls = sp.lambdify(sns_args, tuple(s_calls_nonseq), modules=self.lambdify_func)
            if 'select' in inspect.getsource(sns_calls):
                sns_args.extend(select_args_add)

        self.s_syms = s_syms
        self.calls.s = s_calls
        self.calls.s_args = s_args
        self.calls.sns = sns_calls
        self.calls.sns_args = sns_args

    def generate_jacobians(self, diag_eps=1e-8):
        """
        Generate Jacobians and store to corresponding triplets.

        The internal indices of equations and variables are stored, alongside the lambda functions.

        For example, dg/dy is a sparse matrix whose elements are ``(row, col, val)``, where ``row`` and ``col``
        are the internal indices, and ``val`` is the numerical lambda function. They will be stored to

            row -> self.calls._igy
            col -> self.calls._jgy
            val -> self.calls._vgy

        """
        logger.debug('- Generating Jacobians for %s', self.class_name)

        from sympy import Tuple

        # clear storage
        self.df_syms, self.dg_syms = sp.Matrix([]), sp.Matrix([])
        self.calls.clear_ijv()

        # NOTE: SymPy does not allow getting the derivative of an empty array
        if len(self.g_matrix) > 0:
            self.dg_syms = self.g_matrix.jacobian(self.vars_list)

        if len(self.f_matrix) > 0:
            self.df_syms = self.f_matrix.jacobian(self.vars_list)

        self.df_sparse = sp.SparseMatrix(self.df_syms)
        self.dg_sparse = sp.SparseMatrix(self.dg_syms)

        vars_syms_list = list(self.vars_dict)
        algebs_and_ext_list = list(self.cache.algebs_and_ext)
        states_and_ext_list = list(self.cache.states_and_ext)

        fg_sparse = [self.df_sparse, self.dg_sparse]
        j_args = defaultdict(list)   # argument list for each jacobian call
        j_calls = defaultdict(list)  # jacobian functions (one for each type)

        for idx, eq_sparse in enumerate(fg_sparse):
            for item in eq_sparse.row_list():
                e_idx, v_idx, e_symbolic = item
                if idx == 0:
                    eq_name = states_and_ext_list[e_idx]
                else:
                    eq_name = algebs_and_ext_list[e_idx]

                var_name = vars_syms_list[v_idx]
                eqn = self.cache.all_vars[eq_name]    # `BaseVar` that corr. to the equation
                var = self.cache.all_vars[var_name]   # `BaseVar` that corr. to the variable
                jname = f'{eqn.e_code}{var.v_code}'

                # jac calls with all arguments and stored individually
                self.calls.append_ijv(jname, e_idx, v_idx, 0)

                # collect unique arguments for jac calls
                free_syms = self._check_expr_symbols(e_symbolic)
                for fs in free_syms:
                    if fs not in j_args[jname]:
                        j_args[jname].append(fs)
                j_calls[jname].append(e_symbolic)

        for jname in j_calls:
            # sort for stable argument list
            j_args[jname].sort(key=lambda s: s.name)

            self.calls.j_args[jname] = [str(i) for i in j_args[jname]]
            # workaround for SymPy 1.10 to generate tuples with one element. See
            # https://github.com/sympy/sympy/issues/23224
            self.calls.j[jname] = sp.lambdify(j_args[jname], Tuple(*j_calls[jname]), modules=self.lambdify_func)

            # manually append additional arguments for select
            if 'select' in inspect.getsource(self.calls.j[jname]):
                self.calls.j_args[jname].extend(select_args_add)

        self.calls.j_names = list(j_calls.keys())

        # The for-loop below is intended to add an epsilon small value to the diagonal of `gy`.
        # The user should take care of the algebraic equations by using `diag_eps` in `Algeb` definition

        for var in self.parent.cache.vars_int.values():
            if var.diag_eps == 0.0:
                continue
            elif var.diag_eps is True:
                if self.get_system_config('diag_eps') is not None:
                    eps = self.parent.system.config.diag_eps
                elif diag_eps is not None:
                    eps = diag_eps  # from function argument
                else:
                    eps = 1e-8

            else:
                eps = var.diag_eps

            if var.e_code == 'g':
                eq_list = algebs_and_ext_list
            else:
                eq_list = states_and_ext_list

            e_idx = eq_list.index(var.name)
            v_idx = vars_syms_list.index(var.name)

            self.calls.append_ijv(f'{var.e_code}{var.v_code}c', e_idx, v_idx, eps)

    def generate_pretty_print(self):
        """
        Generate pretty print variables and equations.
        """
        logger.debug("- Generating pretty prints for %s", self.class_name)

        # equation symbols for pretty printing
        self.f, self.g = sp.Matrix([]), sp.Matrix([])

        try:
            self.xy = sp.Matrix(list(self.vars_dict.values())).subs(self.tex_names)
        except TypeError as e:
            logger.error("Error while substituting tex_name for variables.")
            logger.error("Variable names might have conflicts with SymPy functions.")
            raise e

        # get pretty printing equations by substituting symbols
        self.f = self.f_matrix.subs(self.tex_names)
        self.g = self.g_matrix.subs(self.tex_names)
        self.s = [item.subs(self.tex_names) for item in self.s_syms.values()]

        # store latex strings
        nx = len(self.f)
        ny = len(self.g)
        self.calls.x_latex = [sp.latex(item) for item in self.xy[:nx]]
        self.calls.y_latex = [sp.latex(item) for item in self.xy[nx:nx + ny]]

        self.calls.f_latex = [sp.latex(item) for item in self.f]
        self.calls.g_latex = [sp.latex(item) for item in self.g]
        self.calls.s_latex = [sp.latex(item) for item in self.s]

        self.df = self.df_sparse.subs(self.tex_names)
        self.dg = self.dg_sparse.subs(self.tex_names)

        # store init latex strings
        init_latex = OrderedDict()
        for name, instance in self.cache.all_vars.items():
            if instance.v_str is None and instance.v_iter is None:
                init_latex[name] = ''
            else:
                if instance.v_str is not None:
                    init_latex[name] = sp.latex(self.v_str_syms[name].subs(self.tex_names))
                if instance.v_iter is not None:
                    init_latex[name] = sp.latex(self.v_iter_syms[name].subs(self.tex_names))

        self.calls.init_latex = init_latex

    def generate_pycode(self, pycode_path, yapf_pycode):
        """
        Create output source code file for generated code.

        Generated code are stored at ``~/.andes/pycode``.

        Notes
        -----
        In the current implementation, each model saves a ``.py`` file.
        In systems with slow disk access (such as networked file systems),
        this function can be the bottleneck.
        """

        out = list()
        yf = yapf_pycode

        header = \
            """from collections import OrderedDict  # NOQA

from numpy import ones_like, zeros_like, full, array                # NOQA
from numpy import nan, pi, sin, cos, tan, sqrt, exp, select         # NOQA
from numpy import greater_equal, less_equal, greater, less, equal   # NOQA
from numpy import logical_and, logical_or, logical_not              # NOQA
from numpy import real, imag, conj, angle, radians, abs             # NOQA
from numpy import arcsin, arccos, arctan, arctan2                   # NOQA
from numpy import log                                               # NOQA

from andes.thirdparty.npfunc import *                               # NOQA

"""
        # header goes first
        out.append(header)

        # checksum
        out.append(f'md5 = "{self.calls.md5}"\n')

        # equations
        out.append(self._rename_func(self.calls.f, 'f_update', yf))
        out.append(self._rename_func(self.calls.g, 'g_update', yf))

        # jacobians
        for name in self.calls.j:
            out.append(self._rename_func(self.calls.j[name], f'{name}_update', yf))

        # initialization: assignments
        for name in self.calls.ia:
            out.append(self._rename_func(self.calls.ia[name], f'{name}_ia', yf))
        for name in self.calls.ii:
            out.append(self._rename_func(self.calls.ii[name], f'{name}_ii', yf))
        for name in self.calls.ij:
            out.append(self._rename_func(self.calls.ij[name], f'{name}_ij', yf))

        # services
        for name in self.calls.s:
            out.append(self._rename_func(self.calls.s[name], f'{name}_svc', yf))

        # non-sequential services
        out.append(self._rename_func(self.calls.sns, 'sns_update', yf))

        # variables
        for name in dilled_vars:
            out.append(f'{name} = ' + pprint.pformat(self.calls.__dict__[name]))

        out_str = '\n'.join(out)

        pycode_path = get_pycode_path(pycode_path, mkdir=True)
        file_path = os.path.join(pycode_path, f'{self.class_name}.py')

        # do not overwrite file if already up to date
        if os.path.isfile(file_path):
            with open(file_path, 'r') as f:
                fread = f.readlines()

            fread = ''.join(fread)
            if fread == out_str:
                logger.debug("<%s>: generated code is up-to-date and not overwritten.",
                             self.parent.class_name)
                return

        with open(file_path, 'w') as f:
            f.write(out_str)

    def _rename_func(self, func, func_name, yapf_pycode=False):
        """
        Rename the function name and return source code.

        This function performs these tasks:

        1. rename ``_lambdifygenerated`` to the given ``func_name``.
        2. append four arguments if ``select`` is used to pass numba
           compilation.
        3. remove ``Indicator`` for wrappers of logic expressions.

        This function does not check for name conflicts. Install `yapf` for
        optional code reformatting (takes extra processing time).

        It also patches function argument list for select.
        """

        if func is None:
            return f"# empty {func_name}\n"

        src = inspect.getsource(func)
        src = src.replace("def _lambdifygenerated(", f"def {func_name}(")

        # remove `Indicator`
        src = src.replace("Indicator", "")

        # append additional arguments for select
        if 'select' in src:
            right_parenthesis = ", " + ', '.join(select_args_add) + "):"
            src = src.replace("):", right_parenthesis)

        if yapf_pycode:
            try:
                from yapf.yapflib.yapf_api import FormatCode
                src = FormatCode(src, style_config='pep8')[0]  # drop the encoding `None`
            except ImportError:
                logger.warning("`yapf` not installed. Skipped code reformatting.")

        src += '\n'
        return src

    def generate_dependency(self):
        """
        Generate dependency list and initialization order.
        """
        self.v_str_syms = OrderedDict()
        self.v_iter_syms = OrderedDict()
        deps = OrderedDict()

        # convert to symbols
        for name, instance in self.cache.all_vars.items():
            if instance.v_str is not None:
                sympified = sp.sympify(instance.v_str, locals=self.inputs_dict)
                sympified = self._do_substitute(sympified)
                self._check_expr_symbols(sympified)
                self.v_str_syms[name] = sympified
            else:
                # default initial values to zero
                sympified = sp.sympify('0.0', locals=self.inputs_dict)
                self.v_str_syms[name] = sympified

            if instance.v_iter is not None:
                sympified = sp.sympify(instance.v_iter, locals=self.inputs_dict)
                sympified = self._do_substitute(sympified)
                self._check_expr_symbols(sympified)
                self.v_iter_syms[name] = sympified

        # store deps for explicit and iterative initializers
        for name, expr in self.v_str_syms.items():
            _store_deps(name, expr, self.vars_dict, deps)

        for name, expr in self.v_iter_syms.items():
            _store_deps(name, expr, self.vars_dict, deps)

        # store deps for manually added dependent variables
        for name, instance in self.cache.vars_int.items():
            if instance.deps is not None:
                deps[name].extend(instance.deps)

        # resolve dependency
        self.init_seq = resolve_deps(deps)
        self.calls.init_seq = self.init_seq

    def check_v_iter(self):
        """
        Helper function to check if `v_iter` is defined for variables
        with circular dependencies.
        """

        for item in self.init_seq:
            if not isinstance(item, list):
                continue

            for vi in item:
                if self.cache.all_vars[vi].v_iter is None:
                    logger.error("%s: v_iter not defined for %s" % (self.class_name, vi))

    def generate_init(self):
        """
        Generate initialization equations.
        """

        self.generate_dependency()
        self.check_v_iter()
        self.generate_init_eqn()
        self.lambdify_init()

    def generate_init_eqn(self):
        """
        Generate initialization equations.

        The RHS of assignment equations ``v = v_str(x, y)`` or RHS of iterative
        initialization equations in the form of ``0 = v_iter(x, y)`` will be
        stored to ``self.init_list``.

        A list of flags will be stored to ``self.init_flag`` with 0 for
        assignments and 1 for iterative.

        For iteratively initialized variables that require assigned initial
        values (to improve convergence), the initial value can be provided
        through `self.v_str`.
        """

        self.init_asn = OrderedDict()       # assignment-type initialization
        self.init_itn = OrderedDict()       # iterative initialization
        self.init_itn_vars = OrderedDict()  # variables corr. to iterative vars
        self.init_jac = OrderedDict()

        for item in self.init_seq:
            if isinstance(item, str):
                instance = self.parent.__dict__[item]
                if instance.v_str is not None:
                    self.init_asn[item] = self.v_str_syms[item]
                if instance.v_iter is not None:
                    self.init_itn[item] = sp.Matrix([self.v_iter_syms[item]])
                    self.init_itn_vars[item] = [item]

            elif isinstance(item, list):
                name_concat = '_'.join(item)
                eqn_set = sp.Matrix([self.v_iter_syms[name] for name in item])
                self.init_itn[name_concat] = eqn_set
                self.init_itn_vars[name_concat] = item
                for vv in item:
                    instance = self.parent.__dict__[vv]
                    if instance.v_str is not None:
                        self.init_asn[vv] = self.v_str_syms[vv]

        for name, expr in self.init_itn.items():
            vars_iter = OrderedDict()
            for item in self.init_itn_vars[name]:
                vars_iter[item] = self.vars_dict[item]

            self.init_jac[name] = expr.jacobian(list(vars_iter.values()))

    def lambdify_init(self):
        """
        Convert equations and Jacobians to lambda functions.
        """

        init_a = OrderedDict()
        init_i = OrderedDict()
        init_j = OrderedDict()
        ia_args = OrderedDict()  # arguments for assignment init.
        ii_args = OrderedDict()  # arguments for iterative init.
        ij_args = OrderedDict()

        for name, expr in self.init_asn.items():
            fs = self._check_expr_symbols(expr)
            ia_args[name] = [str(i) for i in fs]
            init_a[name] = sp.lambdify(ia_args[name], expr, modules=self.lambdify_func)
            if 'select' in inspect.getsource(init_a[name]):
                ia_args[name].extend(select_args_add)

        for name, expr in self.init_itn.items():
            fs = self._check_expr_symbols(expr)
            ii_args[name] = [str(i) for i in fs]
            init_i[name] = sp.lambdify(ii_args[name], expr, modules=self.lambdify_func)
            if 'select' in inspect.getsource(init_i[name]):
                ii_args[name].extend(select_args_add)

            jexpr = self.init_jac[name]
            fs = self._check_expr_symbols(jexpr)
            ij_args[name] = [str(i) for i in fs]
            init_j[name] = sp.lambdify(ij_args[name], jexpr, modules=self.lambdify_func)
            if 'select' in inspect.getsource(init_j[name]):
                ij_args[name].extend(select_args_add)

        self.calls.ia = init_a
        self.calls.ii = init_i
        self.calls.ij = init_j
        self.calls.ia_args = ia_args
        self.calls.ii_args = ii_args
        self.calls.ij_args = ij_args

    def get_system_config(self, name):
        """
        Helper function for obtaining system config.
        If System is None, return None.
        """

        if hasattr(self.parent.system, 'config'):
            return self.parent.system.config.__dict__[name]

        return None


def _store_deps(name, sympified, vars_int_dict, deps):
    """
    Helper function to store dependencies to a dict.

    Used by ``resolve``.
    """

    deps[name] = []
    free_symbols = sorted(sympified.free_symbols, key=lambda s: s.name)
    for fs in free_symbols:
        if fs not in vars_int_dict.values():
            continue
        if fs not in deps[name]:
            deps[name].append(str(fs))


def resolve_deps(graph):
    """
    Resolve dependency for a dict-based graph using recursion.

    Parameters
    ----------
    graph : dict
        Each key is a variable name, and the corresponding value
        is the list of variables the key depends on.

    Returns
    -------
    list
        A list of initialization sequence. Elements are either a
        variable name or a list of mutually-dependent variables
        that need iterative initialization.
    """

    seq = list()      # sequence after resolving dependency
    visited = list()  # book keeper of the visited nodes

    cflat = list()    # flattened node in circles
    circles = list()  # circles as lists

    def sub_resolve(name, deps, path):
        for item in deps:
            if (item == name) or (item in visited):
                continue

            # undeclared leaf node
            if (item not in graph):
                if item not in seq:
                    seq.append(item)
                continue

            # circular dependency
            if item in path:
                idx1 = path.index(item)
                idx2 = path.index(name)
                mn = min(idx1, idx2)
                mx = max(idx1, idx2)
                cflat.extend(path[mn:mx+1])
                circles.append(path[mn:mx+1])
                continue

            path.append(item)
            sub_resolve(item, graph[item], path)

        # when all dependent nodes are visited
        if name not in visited:
            # if the current node is not in any circle
            if name not in cflat:
                seq.append(name)

            else:
                for cc in circles:
                    if (name in cc) and (cc not in seq):
                        seq.append(cc)

            visited.append(name)

    for name, deps in graph.items():
        path = list()
        sub_resolve(name, deps, path)

    return seq



"""
CuPy solver that requires the ``cupy`` package.
"""

import numpy as np

from andes.shared import cupy
from andes.linsolvers.scipy import SciPySolver


class CuPySolver(SciPySolver):
    """
    CuPy lsqr solver (GPU-based).
    """

    def solve(self, A, b):
        # delayed import for startup speed
        from cupyx.scipy.sparse import csc_matrix as csc_cu  # NOQA
        from cupyx.scipy.sparse.linalg import lsqr as cu_lsqr  # NOQA

        A_csc = self.to_csc(A)

        cu_A = csc_cu(A_csc)
        cu_b = cupy.array(np.array(b).ravel())
        x = cu_lsqr(cu_A, cu_b)

        return np.ravel(cupy.asnumpy(x[0]))



"""
Module for specifying output variables as part of the data file.
"""

import numpy as np
from andes.core.model import ModelData, Model
from andes.core.param import DataParam


class OutputData(ModelData):
    """
    Data for outputs.
    """

    def __init__(self):
        ModelData.__init__(self, three_params=False)

        self.model = DataParam(info='Name of the model', mandatory=True)
        self.varname = DataParam(info='Variable name', )
        self.dev = DataParam(info='Device name', )


class Output(OutputData, Model):
    """
    Model for specifying output models and/or variables.
    """

    def __init__(self, system, config):
        OutputData.__init__(self)
        Model.__init__(self, system, config)

        self.group = 'OutputSelect'
        self.xidx = []
        self.yidx = []

    def in1d(self, addr, v_code):
        """
        Helper function for finding boolean flags to indicate intersections.

        Parameters
        ----------
        idx : array-like
            indices to find
        v_code : str
            variable code in 'x' and 'y'
        """

        if v_code == 'x':
            return np.in1d(self.xidx, addr)
        if v_code == 'y':
            return np.in1d(self.yidx, addr)

        raise NotImplementedError("v_code <%s> not recognized" % v_code)

    def to_output_addr(self, addr, v_code):
        """
        Convert DAE-based variable address to relative output addresses.
        """

        bool_intersect = self.in1d(addr, v_code)
        return np.where(bool_intersect)



"""
Stability criteria module.
"""
import logging
import numpy as np

logger = logging.getLogger(__name__)


def deltadelta(delta, diff_limit):
    """
    Test if a system is stable by comparing  the maximum rotor angle difference
    with a threshold.

    Returns
    -------
    bool
        True if the system is stable, False otherwise.
    """

    if len(delta) < 2:
        return True

    diff_max = np.max(delta - np.min(delta))

    return (diff_max < np.deg2rad(diff_limit)).tolist()



from andes.core.block import Lag, LeadLag, Washout
from andes.core.discrete import HardLimiter
from andes.core.param import NumParam
from andes.core.service import PostInitService
from andes.core.var import Algeb
from andes.models.exciter.excbase import ExcBase, ExcBaseData


class EXST1Data(ExcBaseData):
    """Parameters for EXST1."""

    def __init__(self):
        ExcBaseData.__init__(self)

        self.TR = NumParam(default=0.01,
                           info='Measurement delay',
                           tex_name='T_R',
                           )
        self.VIMAX = NumParam(default=0.2,
                              info='Max. input voltage',
                              tex_name='V_{IMAX}',
                              )

        self.VIMIN = NumParam(default=0,
                              info='Min. input voltage',
                              tex_name='V_{IMIN}',
                              )
        self.TC = NumParam(default=1,
                           info='LL numerator',
                           tex_name='T_C',
                           )
        self.TB = NumParam(default=1,
                           info='LL denominator',
                           tex_name='T_B',
                           )
        self.KA = NumParam(default=80,
                           info='Regulator gain',
                           tex_name='K_A',
                           )
        self.TA = NumParam(default=0.05,
                           info='Regulator delay',
                           tex_name='T_A',
                           )
        self.VRMAX = NumParam(default=8,
                              info='Max. regulator output',
                              tex_name='V_{RMAX}',
                              )

        self.VRMIN = NumParam(default=-3,
                              info='Min. regulator output',
                              tex_name='V_{RMIN}',
                              )
        self.KC = NumParam(default=0.2,
                           info='Coef. for Ifd',
                           tex_name='K_C',
                           )
        self.KF = NumParam(default=0.1,
                           info='Feedback gain',
                           tex_name='K_F',
                           )
        self.TF = NumParam(default=1.0,
                           info='Feedback delay',
                           tex_name='T_F',
                           non_negative=True,
                           non_zero=True,
                           )


class EXST1Model(ExcBase):
    def __init__(self, system, config):
        ExcBase.__init__(self, system, config)

        self.vref = Algeb(info='Reference voltage input',
                          tex_name='V_{ref}',
                          unit='p.u.',
                          v_str='v + vf0 / KA',
                          e_str='vref0 - vref'
                          )

        self.vref0 = PostInitService(info='constant vref',
                                     v_str='vref',
                                     tex_name='V_{ref0}',
                                     )

        # input excitation voltages; PSS outputs summed at vi
        self.vi = Algeb(info='Total input voltages',
                        tex_name='V_i',
                        unit='p.u.',
                        )
        self.vi.v_str = 'vf0 / KA'
        self.vi.e_str = '(vref - LG_y - WF_y) - vi'

        self.LG = Lag(u=self.v, T=self.TR, K=1,
                      info='Sensing delay',
                      )

        self.HLI = HardLimiter(u=self.vi, lower=self.VIMIN, upper=self.VIMAX,
                               info='Hard limiter on input',
                               )

        self.vl = Algeb(info='Input after limiter',
                        tex_name='V_l',
                        v_str='HLI_zi*vi + HLI_zu*VIMAX + HLI_zl*VIMIN',
                        e_str='HLI_zi*vi + HLI_zu*VIMAX + HLI_zl*VIMIN - vl',
                        )

        self.LL = LeadLag(u=self.vl, T1=self.TC, T2=self.TB, info='Lead-lag compensator', zero_out=True)

        self.LR = Lag(u=self.LL_y, T=self.TA, K=self.KA, info='Regulator')

        self.WF = Washout(u=self.LR_y, T=self.TF, K=self.KF, info='Stablizing circuit feedback')

        # the following uses `XadIfd` for `IIFD` in the PSS/E manual
        self.vfmax = Algeb(info='Upper bound of output limiter',
                           tex_name='V_{fmax}',
                           v_str='VRMAX - KC * XadIfd',
                           e_str='VRMAX - KC * XadIfd - vfmax',
                           )
        self.vfmin = Algeb(info='Lower bound of output limiter',
                           tex_name='V_{fmin}',
                           v_str='VRMIN - KC * XadIfd',
                           e_str='VRMIN - KC * XadIfd - vfmin',
                           )

        self.HLR = HardLimiter(u=self.WF_y, lower=self.vfmin, upper=self.vfmax,
                               info='Hard limiter on regulator output')

        self.vout.e_str = 'ue * (LR_y*HLR_zi + vfmin*HLR_zl + vfmax*HLR_zu) - vout'


class EXST1(EXST1Data, EXST1Model):
    """
    EXST1-type static excitation system.
    """

    # TODO: disable offline EXST1 in eigenvalue analysis

    def __init__(self, system, config):
        EXST1Data.__init__(self)
        EXST1Model.__init__(self, system, config)



"""
Tests for PSS/E parser.
"""

import os
import unittest
import yaml

from andes.system import System
from andes.io import psse


class TestPSSEParser(unittest.TestCase):
    """
    Test PSSE parsers.
    """

    def test_sort_models(self):
        """
        Test sort_models.
        """

        system = System()

        dirname = os.path.dirname(__file__)
        with open(f'{dirname}/../andes/io/psse-dyr.yaml', 'r') as f:
            dyr_yaml = yaml.full_load(f)
        psse.sort_psse_models(dyr_yaml, system)



from andes.core.service import VarService
from andes.models.exciter.exdc2 import EXDC2Data, EXDC2Model


class IEEEX1Model(EXDC2Model):

    def __init__(self, system, config):
        EXDC2Model.__init__(self, system, config)
        self.VRTMAX = VarService('VRMAX * v',
                                 tex_name='V_{RMAX}V_T')
        self.VRTMIN = VarService('VRMIN * v',
                                 tex_name='V_{RMIN}V_T')

        self.LA.upper = self.VRTMAX
        self.LA.lower = self.VRTMIN


class IEEEX1(EXDC2Data, IEEEX1Model):
    """
    IEEEX1 Type 1 exciter (DC)

    Derived from EXDC2 by varying the limiter bounds.
    """
    def __init__(self, system, config):
        EXDC2Data.__init__(self)
        IEEEX1Model.__init__(self, system, config)

        self.vout.e_str = 'ue * vp - vout'



"""
Interoperability with MATPOWER through ``matpower-pip``.

Please install the Python package ``matpower`` and configure MATLAB or Octave,
following the instructions at matpower-pip_.

To create a MATLAB/Octave instance, do:

.. code:: python

    from andes.interop.matpower import start_instance

    m = start_instance()

.. _matpower-pip: https://github.com/yasirroni/matpower-pip

"""


import logging
from functools import wraps

import andes

from andes.shared import Oct2PyError

logger = logging.getLogger(__name__)


def require_matpower(f):
    """
    Decorator for functions that require matpower.
    """

    @wraps(f)
    def wrapper(*args, **kwds):
        try:
            from matpower import start_instance   # NOQA
        except ImportError:
            raise ModuleNotFoundError("Package `matpower` needs to be manually installed.")

        return f(*args, **kwds)

    return wrapper


@require_matpower
def from_matpower(m, varname, system=None):
    """
    Retrieve a MATPOWER mpc case from a MATLAB/Octave instance.

    Parameters
    ----------
    m : MATLAB/Octave instance
        Instance from which to retrieve the MATPOWER case.
    varname : str
        Name of the variable in the MATPOWER MPC format to retrieve.

    Returns
    -------
    :class:`~andes.system.System`
        System from the mpc case. The system will not have been set up.

    Examples
    --------
    To retrieve a case from MATPOWER from instance ``m``, do the following:

    .. code:: python


        from andes.interop.matpower import start_instance, to_matpower, from_matpower

        system = from_matpower(m, 'mpc')

    One can create an Excel file with dynamic data only and use the ``xlsx``
    parser to load data into ``system``:

    .. code:: python

        from andes.io import xlsx

        xlsx.read(system, andes.get_case('ieee14/ieee14_dyn_only.xlsx'))

    The ``ieee14_dyn_only.xlsx`` is an example spreadsheet that only contains
    dynamic components. One will need to create the ``idx`` correctly for
    dynamic components to match these from the MATPOWER case. If not sure about
    the indices, one can save the ANDES system to an Excel file, using:

    .. code:: python

        xlsx.write(system, 'system_static.xlsx')

    """

    if m is None:
        raise ImportError("MATPOWER is not installed or not properly configured.")

    mpc = {}
    mpc['baseMVA'] = m.eval(f'{varname}.baseMVA;')
    mpc['version'] = m.eval(f'{varname}.version;')
    mpc['bus'] = m.eval(f'{varname}.bus;')
    mpc['gen'] = m.eval(f'{varname}.gen;')
    mpc['branch'] = m.eval(f'{varname}.branch;')

    try:
        mpc['gencost'] = m.eval(f'{varname}.gencost;')
    except Oct2PyError:
        logger.debug("No gencost in mpc")
    try:
        mpc['bus_names'] = m.eval(f'{varname}.bus_names;')
    except Oct2PyError:
        logger.debug("No bus_name in mpc")

    if system is None:
        system = andes.System()

    andes.io.matpower.mpc2system(mpc, system)

    return system


@require_matpower
def to_matpower(m, varname, system):
    """
    Send an ANDES case to a running MATLAB instance.

    Parameters
    ----------
    m : MATLAB/Octave instance
        Instance to which to send the MATPOWER case.
    varname : str
        Name of the variable to store the mpc case in MATLAB/Octave.
    system : :class:`~andes.system.System`
        System whose power flow data to send to MATPOWER.

    Examples
    --------

    The code below will create an IEEE 14-bus example system in ANDES, convert it to
    MATPOWER's case, and send to the MATLAB/Octave instance.

    .. code:: python

        import andes

        from andes.interop.matpower import (start_instance,
            to_matpower, from_matpower)

        m = start_instance()

        ss = andes.system.example()
        mpc = to_matpower(m, 'mpc', ss)

        m.eval("runpf(mpc)")

        mpc_out = m.pull("mpc")  # retrieve the mpc case from MATLAB/Octave

    """

    if m is None:
        raise ImportError("MATPOWER is not installed or not properly configured.")

    mpc = andes.io.matpower.system2mpc(system)
    m.push(varname, mpc)

    return mpc



"""
Test the ANDES-pandapower interface.
"""

import unittest
import numpy as np

import andes
from andes.shared import deg2rad
from andes.interop.pandapower import to_pandapower, make_link_table, make_GSF

try:
    import pandapower as pp
    getattr(pp, '__version__')
    HAVE_PANDAPOWER = True
except (ImportError, AttributeError):
    HAVE_PANDAPOWER = False


@unittest.skipUnless(HAVE_PANDAPOWER, "pandapower not available")
class TestPandapower(unittest.TestCase):
    """
    Tests for the ANDES-pandapower interface.
    """

    cases = ['ieee14/ieee14_ieeet1.xlsx',
             'ieee14/ieee14_pvd1.xlsx',
             'ieee39/ieee39.xlsx',
             'npcc/npcc.xlsx',
             ]

    def setUp(self) -> None:
        """
        Test setup. This is executed before each test case.
        """

    def test_to_pandapower(self):
        """
        Test `andes.interop.pandapower.to_pandapower` with cases
        """
        for case_file in self.cases:
            case = andes.get_case(case_file)

            _test_to_pandapower_single(case, tol=1e-3)

    def test_make_link_table(self):
        """
        Test `andes.interop.pandapower.make_link_table`
        """

        sa14 = andes.load(andes.get_case('ieee14/ieee14_ieeet1.xlsx'),
                          setup=True,
                          no_output=True,
                          default_config=True)
        link_table = make_link_table(sa14)
        ridx = link_table[link_table['syg_idx'] == 'GENROU_1'].index
        c_bus = link_table['bus_name'].iloc[ridx].astype(str) == 'BUS1'
        c_exc = link_table['exc_idx'].iloc[ridx].astype(str) == 'ESST3A_2'
        c_stg = link_table['stg_idx'].iloc[ridx].astype(str) == '1'
        c_gov = link_table['gov_idx'].iloc[ridx].astype(str) == 'TGOV1_1'
        c = c_bus.values[0] and c_exc.values[0] and c_stg.values[0] and c_gov.values[0]
        return c

    def test_make_GSF(self):
        """
        Test `andes.interop.pandapower.make_GSF with ieee39`
        """

        sa39 = andes.load(andes.get_case('ieee39/ieee39.xlsx'),
                          setup=True,
                          no_output=True,
                          default_config=True)
        sp39 = to_pandapower(sa39)
        make_GSF(sp39)
        return True


def _test_to_pandapower_single(case, **kwargs):
    """
    Test `andes.interop.pandapower.to_pandapower` with a single case
    """

    sa = andes.load(case, setup=True, no_output=True, default_config=True)
    sp = to_pandapower(sa, **kwargs)

    sa.PFlow.run()
    pp.runpp(sp)

    v_andes = sa.Bus.v.v
    a_andes = sa.Bus.a.v
    v_pp = sp.res_bus['vm_pu']
    a_pp = sp.res_bus['va_degree'] * deg2rad

    # align ssa angle with slcka bus angle
    rid_slack = np.argmin(np.abs(a_pp))
    a_andes = a_andes - a_andes[rid_slack]

    return np.testing.assert_almost_equal(v_andes, v_pp, decimal=3)



"""
Module for ModelCall.
"""

from collections import OrderedDict

from andes.shared import jac_names, jac_types


class ModelCall:
    """
    Class for storing generated function calls, Jacobian calls, and arguments.
    """

    def __init__(self):
        self.md5 = ''

        # `f` and `g` are callables generated by lambdify that take positional args
        self.f = None
        self.g = None
        self.j = dict()
        self.s = OrderedDict()
        self.sns = None

        # `f_args` and `g_args` are the arg names
        self.f_args = list()
        self.g_args = list()
        self.j_args = dict()
        self.ia = OrderedDict()
        self.ii = OrderedDict()
        self.ij = OrderedDict()
        self.s_args = OrderedDict()
        self.sns_args = list()
        self.ia_args = OrderedDict()  # assignment initialization
        self.ii_args = OrderedDict()  # iterative initialization
        self.ij_args = OrderedDict()

        self.ijac = OrderedDict()
        self.jjac = OrderedDict()
        self.vjac = OrderedDict()

    def clear_ijv(self):
        for jname in jac_names:
            for jtype in jac_types:
                self.ijac[jname + jtype] = list()
                self.jjac[jname + jtype] = list()
                self.vjac[jname + jtype] = list()

    def append_ijv(self, j_full_name, ii, jj, vv):
        if not isinstance(ii, int):
            raise ValueError("i index must be an integer")
        if not isinstance(jj, int):
            raise ValueError("j index must be an integer")
        if not isinstance(vv, (int, float)) and (not callable(vv)):
            raise ValueError("v must be a number or a callable")

        self.ijac[j_full_name].append(ii)
        self.jjac[j_full_name].append(jj)
        self.vjac[j_full_name].append(vv)

    def zip_ijv(self, j_full_name):
        """
        Return a zipped iterator for the rows, cols and vals for the specified matrix name.
        """
        return zip(self.ijac[j_full_name],
                   self.jjac[j_full_name],
                   self.vjac[j_full_name])



from andes.core.block import Lag, LeadLag
from andes.core.discrete import HardLimiter
from andes.core.param import NumParam
from andes.core.service import PostInitService
from andes.core.var import Algeb
from andes.models.exciter.excbase import ExcBase, ExcBaseData


class EXAC4Data(ExcBaseData):
    def __init__(self):
        ExcBaseData.__init__(self)
        self.TR = NumParam(info='Sensing time constant',
                           tex_name='T_R',
                           default=0.01,
                           unit='p.u.',
                           )
        self.VIMAX = NumParam(default=5.0,
                              info='Max. input voltage',
                              tex_name='V_{IMAX}',
                              vrange=(0, 5),
                              )
        self.VIMIN = NumParam(default=-0.1,
                              info='Min. input voltage',
                              tex_name='V_{IMIN}',
                              vrange=(-1, 0),
                              )
        self.TC = NumParam(info='Lead time constant in lead-lag',
                           tex_name='T_C',
                           default=1,
                           unit='p.u.',
                           )
        self.TB = NumParam(info='Lag time constant in lead-lag',
                           tex_name='T_B',
                           default=1,
                           unit='p.u.',
                           )
        self.KA = NumParam(default=80,
                           info='Regulator gain',
                           tex_name='K_A',
                           )
        self.TA = NumParam(info='Lag time constant in regulator',
                           tex_name='T_A',
                           default=0.04,
                           unit='p.u.',
                           )
        self.VRMAX = NumParam(info='Maximum excitation limit',
                              tex_name='V_{RMAX}',
                              default=8,
                              unit='p.u.',
                              vrange=(0.5, 10),
                              )
        self.VRMIN = NumParam(info='Minimum excitation limit',
                              tex_name='V_{RMIN}',
                              default=0,
                              unit='p.u.',
                              vrange=(-10, 0.5),
                              )
        self.KC = NumParam(default=0.0,
                           tex_name='K_C',
                           info='Reactive power compensation gain',
                           )


class EXAC4Model(ExcBase):
    # TODO: check why `vref` is not used
    def __init__(self, system, config):
        ExcBase.__init__(self, system, config)

        self.LG = Lag(u=self.v, T=self.TR, K=1,
                      info='Sensing delay',
                      )
        self.vi = Algeb(info='Total input voltages',
                        tex_name='V_i',
                        unit='p.u.',
                        )
        self.vi.v_str = 'vf0 / KA'
        self.vi.e_str = '(vref0 - LG_y) - vi'

        self.vref0 = PostInitService(info='Const reference voltage',
                                     tex_name='V_{ref0}',
                                     v_str='v + vf0 / KA',
                                     )

        self.HLI = HardLimiter(u=self.vi, lower=self.VIMIN, upper=self.VIMAX,
                               info='Hard limiter on input',
                               )

        self.LL = LeadLag(u='vi * HLI_zi + VIMIN * HLI_zl + VIMAX * HLI_zu',
                          T1=self.TC,
                          T2=self.TB,
                          info='Lead-lag compensator',
                          zero_out=True,
                          )

        self.LR = Lag(u=self.LL_y, T=self.TA, K=self.KA, info='Regulator')

        # the following uses `XadIfd` for `IIFD` in the PSS/E manual
        self.vfmax = Algeb(info='Upper bound of output limiter',
                           tex_name='V_{fmax}',
                           v_str='VRMAX - KC * XadIfd',
                           e_str='VRMAX - KC * XadIfd - vfmax',
                           )
        self.vfmin = Algeb(info='Lower bound of output limiter',
                           tex_name='V_{fmin}',
                           v_str='VRMIN - KC * XadIfd',
                           e_str='VRMIN - KC * XadIfd - vfmin',
                           )

        self.HLR = HardLimiter(u=self.LR_y, lower=self.vfmin, upper=self.vfmax,
                               info='Hard limiter on regulator output')

        self.vout.e_str = 'ue * (LR_y*HLR_zi + vfmin*HLR_zl + vfmax*HLR_zu) - vout'


class EXAC4(EXAC4Data, EXAC4Model):
    """
    IEEE Type AC4 excitation system model.
    """

    def __init__(self, system, config):
        EXAC4Data.__init__(self)
        EXAC4Model.__init__(self, system, config)
